from quicktions import Fraction

from musicscore.musictree.treepart import TreePart
from musicscore.musictree.treetime import TreeTime
from musicscore.musicxml.elements import timewise as timewise
from musicscore.musicxml.elements.barline import Barline, BarStyle
from musicscore.musicxml.groups.layout import SystemLayout
from musicscore.musicxml.groups.musicdata import Print, Attributes
from musicscore.musicxml.types.complextypes.systemlayout import SystemDistance


class TreeMeasure(timewise.Measure):
    """"""

    def __init__(self, time=(4, 4), *args, **kwargs):
        super().__init__(number='1', *args, **kwargs)
        self._time = None
        self.time = time
        self._beats = None
        self._offset = None
        self._barline_style = None

    @property
    def __name__(self):
        index = self.up.get_children_by_type(self.__class__).index(self)
        return str(index + 1)

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, value):
        if value:
            if isinstance(value, TreeTime):
                self._time = value
            else:
                if isinstance(self._time, TreeTime):
                    self._time.reset_dtd()
                    self._time.pars_arguments(value)
                else:
                    self._time = TreeTime(*value)
        else:
            self._time = None

    @property
    def quarter_duration(self):
        output = 0
        for time_signature in self.time.get_time_signatures():
            (beats, beat_type) = time_signature
            output += beats.value / beat_type.value * 4
        return Fraction(output).limit_denominator(10000)

    @property
    def barline_style(self):
        return self._barline_style

    def set_barline_style(self, value):
        if value:
            for part in self.get_children_by_type(TreePart):
                try:
                    bl = part.get_children_by_type(Barline)[0]
                except IndexError:
                    bl = part.add_child(Barline())

                try:
                    bs = bl.get_children_by_type(BarStyle)[0]
                    bs.value = value
                except IndexError:
                    bs = bl.add_child(BarStyle(value))

        self._barline_style = value

    def show_time_signature(self):
        for part in self.get_children_by_type(TreePart):
            part.get_children_by_type(Attributes)[0].add_child(self.time)

    def hide_time_signature(self):
        part = self.get_children_by_type(TreePart)[0]
        if self.time in part.get_children_by_type(Attributes)[0].get_children():
            part.get_children_by_type(Attributes)[0].remove_child(self.time)

    def get_part(self, number):
        return self.get_children_by_type(TreePart)[number - 1]

    def get_beats(self):
        output = []
        for part in self.get_children_by_type(TreePart):
            output.extend(part.get_beats())
        return output

    @property
    def previous(self):
        if not self.up:
            return None
        index = self.up.get_children_by_type(TreeMeasure).index(self)
        if index == 0:
            return None
        return self.up.get_children_by_type(TreeMeasure)[index - 1]

    def update_offset(self):
        if self.previous:
            output = self.previous.offset + self.previous.quarter_duration
            self._offset = output
        else:
            self._offset = 0

    def _add_page_one(self):
        score = self.up

        first_part = score.get_measure(1).get_part(1)

        first_part.get_children_by_type(TreePart)
        try:
            print_ = first_part.get_children_by_type(Print)[0]
        except IndexError:
            print_ = first_part.add_child(Print())
        print_.page_number = '1'

    def add_page_break(self):

        score = self.up

        if not score._identifications_added:
            score._add_identifications()

        self._add_page_one()

        for part in self.get_children_by_type(TreePart):
            try:
                p = part.get_children_by_type(Print)[0]
            except IndexError:
                p = part.add_child(Print())
            p.new_page = 'yes'

    def add_system_break(self):
        score = self.up
        if not score._identifications_added:
            score._add_identifications()

        for part in self.get_children_by_type(TreePart):
            try:
                p = part.get_children_by_type(Print)[0]
            except IndexError:
                p = part.add_child(Print())
            p.new_system = 'yes'

    def add_system_distance(self, value):
        for part in self.get_children_by_type(TreePart):
            try:
                p = part.get_children_by_type(Print)[0]
            except IndexError:
                p = part.add_child(Print())

            try:
                s = p.get_children_by_type(SystemLayout)[0]
            except IndexError:
                s = p.add_child(SystemLayout())

            try:
                sd = s.get_children_by_type(SystemDistance)[0]
            except IndexError:
                sd = s.add_child(SystemDistance())

            sd.value = value

    @property
    def offset(self):
        if self._offset is None:
            self.update_offset()
        return self._offset

    def __copy__(self):
        time = self.time.__copy__()
        new_measure = TreeMeasure(time)
        for key, new_key in zip(self.__dict__.keys(), new_measure.__dict__.keys()):
            item = self.__dict__[key]
            if key == '_attributes':
                new_measure.__dict__[new_key] = item
        return new_measure



"""
	<xs:complexType name="time-modification">
		<xs:sequence>
			<xs:element name="actual-notes" type="xs:nonNegativeInteger">
				<xs:annotation>
					<xs:documentation></xs:documentation>
				</xs:annotation>
			</xs:element>
			<xs:element name="normal-notes" type="xs:nonNegativeInteger">
				<xs:annotation>
					<xs:documentation>.</xs:documentation>
				</xs:annotation>
			</xs:element>
			<xs:sequence minOccurs="0">
				<xs:element name="normal-type" type="note-type-value">
					<xs:annotation>
						<xs:documentation></xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element name="normal-dot" type="empty" minOccurs="0" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation></xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:sequence>
	</xs:complexType>
"""
from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType, Empty
from musicscore.musicxml.types.simple_type import NonNegativeInteger, TypeNoteTypeValue


class ActualNotes(XMLElement, NonNegativeInteger):
    """
    The actual-notes element describes how many notes are played in the time usually occupied by the number in the
    normal-notes element.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='actual-notes', value=value, *args, **kwargs)


class NormalNotes(XMLElement, NonNegativeInteger):
    """
    The normal-notes element describes how many notes are usually played in the time occupied by the number in the
    actual-notes element
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='normal-notes', value=value, *args, **kwargs)


class NormalType(XMLElement, TypeNoteTypeValue):
    """
    If the type associated with the number in the normal-notes element is different than the current note type (e.g.,
    a quarter note within an eighth note triplet), then the normal-notes type (e.g. eighth) is specified in the
    normal-type and normal-dot elements.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='normal-type', value=value, *args, **kwargs)


class NormalDot(Empty):
    """
    The normal-dot element is used to specify dotted normal tuplet types..
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='normal-dot', value=value, *args, **kwargs)


class ComplexTypeTimeModification(ComplexType):
    """
    Time modification indicates tuplets, double-note tremolos, and other durational changes. A time-modification element
    shows how the cumulative, sounding effect of tuplets and double-note tremolos compare to the written note type
    represented by the type and dot elements. Nested tuplets and other notations that use more detailed information need
    both the time-modification and tuplet elements to be represented accurately.
    """
    _DTD = Sequence(
        Element(ActualNotes),
        Element(NormalNotes),
        Sequence(
            Element(NormalType),
            Element(NormalDot, min_occurrence=0, max_occurrence=None),
            min_occurrence=0
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(tag='time-modification', *args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Rotation(AttributeAbstract):
    """"""

    def __init__(self, rotation=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('rotation', rotation, "TypeRotationDegrees")


class TextRotation(Rotation):
    """
    The rotation attribute is used to rotate text around the alignment point specified by the halign and valign
    attributes. Positive values are clockwise rotations, while negative values are counter-clockwise rotations.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from unittest import TestCase

from musicscore.musictree.treechord import TreeChord
from musicscore.musicxml.elements.fullnote import Pitch, Step


class Test(TestCase):
    def setUp(self) -> None:
        self.chord = TreeChord()

    def test_1(self):
        self.chord.add_midi(60)
        steps = [step.value for note in self.chord._notes for pitch in note.get_children_by_type(Pitch) for step in pitch.get_children_by_type(Step)]
        self.assertEqual(steps, ['B', 'C'])



from unittest import TestCase

from musicscore.musicxml.types.complextypes.encoding import Supports
from musicscore.musicxml.types.complextypes.identification import Encoding
from musicscore.musicxml.types.complextypes.scorepart import Identification


class Test(TestCase):
    def test_1(self):
        identification = Identification()
        encoding = identification.add_child(Encoding())
        supports = encoding.add_child(Supports(type_='yes', element='print', value_='yes'))
        supports.attribute = 'new-page'

        result = '''<identification>
  <encoding>
    <supports attribute="new-page" element="print" type="yes" value="yes"/>
  </encoding>
</identification>
'''
        self.assertEqual(identification.to_string(), result)



from musicscore.dtd.dtd import Choice, Element
from musicscore.musicxml.attributes.enclosure import Enclosure
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printstyle import PrintStyleAlign
from musicscore.musicxml.attributes.textdecoration import TextDecoration
from musicscore.musicxml.types.complextypes.complextype import ComplexType, Empty
from musicscore.musicxml.types.complextypes.othertext import ComplexTypeOtherText


class P(Empty):
    _TAG = 'p'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class PP(Empty):
    _TAG = 'pp'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class PPP(Empty):
    _TAG = 'ppp'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class PPPP(Empty):
    _TAG = 'pppp'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class PPPPP(Empty):
    _TAG = 'ppppp'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class PPPPPP(Empty):
    _TAG = 'pppppp'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class F(Empty):
    _TAG = 'f'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class FF(Empty):
    _TAG = 'ff'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class FFF(Empty):
    _TAG = 'fff'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class FFFF(Empty):
    _TAG = 'ffff'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class FFFFF(Empty):
    _TAG = 'fffff'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class FFFFFF(Empty):
    _TAG = 'ffffff'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class MP(Empty):
    _TAG = 'mp'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class MF(Empty):
    _TAG = 'mf'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SF(Empty):
    _TAG = 'sf'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SFP(Empty):
    _TAG = 'sfp'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SFPP(Empty):
    _TAG = 'sfpp'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class FP(Empty):
    _TAG = 'fp'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class RF(Empty):
    _TAG = 'rf'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class RFZ(Empty):
    _TAG = 'rfz'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SFZ(Empty):
    _TAG = 'sfz'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SFFZ(Empty):
    _TAG = 'sffz'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class FZ(Empty):
    _TAG = 'fz'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class N(Empty):
    _TAG = 'n'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class PF(Empty):
    _TAG = 'pf'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SFZP(Empty):
    _TAG = 'sfzp'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class OtherDynamics(ComplexTypeOtherText):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='other_dynamics', *args, **kwargs)


class ComplexTypeDynamics(ComplexType, PrintStyleAlign, Placement, TextDecoration, Enclosure, OptionalUniqueId):
    """
    Dynamics can be associated either with a note or a general musical direction. To avoid inconsistencies between and
    amongst the letter abbreviations for dynamics (what is sf vs. sfz, standing alone or with a trailing dynamic that
    is not always piano), we use the actual letters as the names of these dynamic elements. The other-dynamics element
    allows other dynamic marks that are not covered here, but many of those should perhaps be included in a more general
    musical direction element. Dynamics elements may also be combined to create marks not covered by a single element,
    such as sfmp.

    These letter dynamic symbols are separated from crescendo, decrescendo, and wedge indications. Dynamic
    representation is inconsistent in scores. Many things are assumed by the composer and left out, such as returns to
    original dynamics. Systematic representations are quite complex: for example, Humdrum has at least 3 representation
    formats related to dynamics. The MusicXML format captures what is in the score, but does not try to be optimal for
    analysis or synthesis of dynamics.
    """

    _DTD = Choice(
        Element(P),
        Element(PP),
        Element(PPP),
        Element(PPPP),
        Element(PPPPP),
        Element(PPPPPP),
        Element(F),
        Element(FF),
        Element(FFF),
        Element(FFFF),
        Element(FFFFF),
        Element(FFFFFF),
        Element(MP),
        Element(MF),
        Element(SF),
        Element(SFP),
        Element(SFPP),
        Element(FP),
        Element(RF),
        Element(RFZ),
        Element(SFZ),
        Element(SFFZ),
        Element(FZ),
        Element(N),
        Element(PF),
        Element(SFZP),
        Element(OtherDynamics),
        min_occurrence=0,
        max_occurrence=None
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.placement = 'below'


class Dynamics(ComplexTypeDynamics):
    _TAG = 'dynamics'
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


from unittest import TestCase

from musicscore.musictree.midi import Midi


class Test(TestCase):
    def test_1(self):
        midi = Midi(60)
        expected = '''<pitch>
  <step>C</step>
  <octave>4</octave>
</pitch>
'''
        self.assertEqual(expected, midi.get_pitch_rest().to_string())

    def test_2(self):
        midi = Midi(0)
        expected = '''<rest/>
'''
        self.assertEqual(expected, midi.get_pitch_rest().to_string())

    def test_3(self):
        midi = Midi(59, accidental_mode='enharmonic_1')
        expected = '''<pitch>
  <step>C</step>
  <alter>-1</alter>
  <octave>4</octave>
</pitch>
'''
        self.assertEqual(expected, midi.get_pitch_rest().to_string())

    def test_4(self):
        midi = Midi(59, accidental_mode='enharmonic_2')
        expected = '''<pitch>
  <step>A</step>
  <alter>2</alter>
  <octave>3</octave>
</pitch>
'''
        self.assertEqual(expected, midi.get_pitch_rest().to_string())

    def test_5(self):
        midi = Midi(58, accidental_mode='sharp')
        expected = '''<pitch>
  <step>A</step>
  <alter>1</alter>
  <octave>3</octave>
</pitch>
'''
        self.assertEqual(expected, midi.get_pitch_rest().to_string())

    def test_6(self):
        midi = Midi(59)
        expected = 'B3'
        self.assertEqual(expected, midi.__name__)

    def test_7(self):
        midi = Midi(58)
        expected = 'Bb3'
        self.assertEqual(expected, midi.__name__)

    def test_8(self):
        midi = Midi(61)
        expected = 'C#4'
        self.assertEqual(expected, midi.__name__)

    def test_9(self):
        midi = Midi(61.5)
        expected = 'D-4'
        self.assertEqual(expected, midi.__name__)

    def test_10(self):
        midi = Midi(61.5, accidental_mode='sharp')
        expected = 'C#+4'
        self.assertEqual(expected, midi.__name__)

    def test_11(self):
        midi = Midi(60.5, accidental_mode='sharp')
        expected = 'C+4'
        self.assertEqual(expected, midi.__name__)

    def test_12(self):
        midi = Midi(60.5, accidental_mode='flat')
        expected = 'Db-4'
        self.assertEqual(expected, midi.__name__)



from unittest import TestCase


class TestScore(TestCase):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def assert_template(self, result_path, template_path=None):
        if not template_path:
            if result_path[-4:] == '.xml':
                template_path = result_path[:-4] + '_template.xml'
            else:
                template_path = result_path + '_template.xml'
        with open(template_path, 'r') as myfile:
            template = myfile.read()
        if result_path[-4:] != '.xml':
            result_path = result_path + '.xml'
        with open(result_path, 'r') as myfile:
            result = myfile.read()
        self.assertEqual(template, result)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treeinstruments import Violin
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()
        sf = SimpleFormat(durations=[4])
        sf.to_stream_voice().add_to_score(self.score)

    def test_1(self):
        self.score.get_score_parts()[0].instrument = Violin()
        self.score.get_score_parts()[0].instrument.part_name.font_size = 9

        xml_path = path + '_test_1.xml'
        self.score.write(path=xml_path)
        # TestScore().assert_template(result_path=xml_path)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Parentheses(AttributeAbstract):

    def __init__(self, parentheses=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('parentheses', parentheses, "TypeYesNo")


class Bracket(AttributeAbstract):

    def __init__(self, bracket=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('bracket', bracket, "TypeYesNo")


class Size(AttributeAbstract):

    def __init__(self, size=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('size', size, "TypeSymbolSize")


class LevelDisplay(Parentheses, Bracket, Size):
    """
    The level-display attribute group specifies three common ways to indicate editorial indications: putting parentheses
    or square brackets around a symbol, or making the symbol a different size. If not specified, they are left to
    application defaults. It is used by the level and accidental elements
    """

    def __init__(self, parentheses=None, bracket=None, size=None, *args, **kwargs):
        super().__init__(parentheses=parentheses, bracket=bracket, size=size, *args, **kwargs)



from musicscore.musicxml.attributes.align import Halign, Valign
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.font import Font
from musicscore.musicxml.attributes.position import Position


class PrintStyle(Position, Font, Color):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class PrintStyleAlign(PrintStyle, Halign, Valign):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from unittest import TestCase
import os

from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        self.score.add_measure()
        self.score.add_part()
        self.score.add_chord(1, 1, TreeChord())
        self.score.add_chord(1, 1, TreeChord(quarter_duration=0))
        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)



from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeBreathMarkValue


class ComplexTypeBreathMark(ComplexType, TypeBreathMarkValue, PrintStyle, Placement):
    """The breath-mark element indicates a place to take a breath."""

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



"""
   <xs:complexType name="tremolo">
        <xs:simpleContent>
            <xs:extension base="tremolo-marks">
                <xs:attribute name="type" type="tremolo-type" default="single"/>
            </xs:extension>
        </xs:simpleContent>
    </xs:complexType>
"""
from musicscore.musicxml.attributes.accidental import Smulf
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeTremoloType, TypeTremoloMarks


class ComplexTypeTremolo(ComplexType, TypeTremoloMarks, PrintStyle, Placement, Smulf):
    """
    The tremolo ornament can be used to indicate single-note, double-note, or unmeasured tremolos. Single-note tremolos
    use the single type, double-note tremolos use the start and stop types, and unmeasured tremolos use the unmeasured
    type. The default is "single" for compatibility with Version 1.1. The text of the element indicates the number of
    tremolo marks and is an integer from 0 to 8. Note that the number of attached beams is not included in this value,
    but is represented separately using the beam element. The value should be 0 for unmeasured tremolos.

    When using double-note tremolos, the duration of each note in the tremolo should correspond to half of the notated
    type value. A time-modification element should also be added with an actual-notes value of 2 and a normal-notes
    value of 1. If used within a tuplet, this 2/1 ratio should be multiplied by the existing tuplet ratio.

    The smufl attribute specifies the glyph to use from the SMuFL tremolos range for an unmeasured tremolo. It is
    ignored for other tremolo types. The SMuFL buzzRoll glyph is used by default if the attribute is missing.

    Using repeater beams for indicating tremolos is deprecated as of MusicXML 3.0.
    """

    def __init__(self, tag, type='single', *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.type = type

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeTremoloType(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)



from musicscore.dtd.dtd import Sequence, Choice, Element
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printobject import PrintObject
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.types.complextypes.complextype import ComplexType, Empty


class Natural(Empty):
    """
    The natural element indicates that this is a natural harmonic. These are usually notated at base pitch rather
    than sounding pitch.
    """
    _TAG = 'natural'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Artificial(Empty):
    """
    The artificial element indicates that this is an artificial harmonic.
    """
    _TAG = 'natural'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class BasePitch(Empty):
    """
    The base pitch is the pitch at which the string is played before touching to create the harmonic.
    """
    _TAG = 'base-pitch'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class TouchingPitch(Empty):
    """
    The touching-pitch is the pitch at which the string is touched lightly to produce the harmonic.
    """
    _TAG = 'touching-pitch'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SoundingPitch(Empty):
    """
    The sounding-pitch is the pitch which is heard when playing the harmonic.
    """
    _TAG = 'sounding-pitch'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeHarmonic(ComplexType, PrintObject, PrintStyle, Placement):
    """
    The harmonic type indicates natural and artificial harmonics. Allowing the type of pitch to be specified, combined
    with controls for appearance/playback differences, allows both the notation and the sound to be represented.
    Artificial harmonics can add a notated touching-pitch; artificial pinch harmonics will usually not notate a touching
    pitch. The attributes for the harmonic element refer to the use of the circular harmonic symbol, typically but not
    always used with natural harmonics.
    """
    _DTD = Sequence(
        Choice(
            Element(Natural),
            Element(Artificial),
            min_occurrence=0
        ),
        Choice(
            Element(BasePitch),
            Element(TouchingPitch),
            Element(SoundingPitch),
            min_occurrence=0
        )

    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



# from musicscore.musictree.treechord import TreeChord
# from musicscore.musictree.treescoretimewise import TreeScoreTimewise
# import os
# import cProfile
#
# path = os.path.abspath(__file__).split('.')[0]
#
#
# def p():
#     score = TreeScoreTimewise()
#     score.add_part('one')
#
#     # midis = [(60, 61)]*500
#     midis = [60]*500
#
#     for index, midi in enumerate(midis):
#         if index % 8 == 0:
#             score.add_measure()
#         score.add_chord((index//8 + 1), 1, TreeChord(midi, quarter_duration=0.5))
#
#     score.finish()
#     score.write(path=path)
#
#
# cProfile.run('p()', sort="tottime")



'''
    <xs:complexType name="system-layout">
        <xs:sequence>
            <xs:element name="system-margins" type="system-margins" minOccurs="0"/>
            <xs:element name="system-dividers" type="system-dividers" minOccurs="0"/>
        </xs:sequence>
    </xs:complexType>
'''
from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.systemdividers import ComplexTypeSystemDividers
from musicscore.musicxml.types.complextypes.systemmargins import ComplexTypeSystemMargins
from musicscore.musicxml.types.simple_type import TypeTenths


class SystemMargins(ComplexTypeSystemMargins):
    _TAG = 'system-margins'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SystemDistance(XMLElement, TypeTenths):
    _TAG = 'system-distance'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class TopSystemDistance(XMLElement, TypeTenths):
    _TAG = 'top-system-distance'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class SystemDividers(ComplexTypeSystemDividers):
    _TAG = 'system-dividers'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeSystemLayout(ComplexType):
    """A system is a group of staves that are read and played simultaneously. System layout includes left and right
    margins and the vertical distance from the previous system. The system distance is measured from the bottom line
    of the previous system to the top line of the current system. It is ignored for the first system on a page. The top
    system distance is measured from the page's top margin to the top line of the first system. It is ignored for all
    but the first system on a page.

    Sometimes the sum of measure widths in a system may not equal the system width specified by the layout elements due
    to roundoff or other errors. The behavior when reading MusicXML files in these cases is application-dependent. For
    instance, applications may find that the system layout data is more reliable than the sum of the measure widths,
    and adjust the measure widths accordingly."""

    _DTD = Sequence(
        Element(SystemMargins, min_occurrence=0),
        Element(SystemDistance, min_occurrence=0),
        Element(TopSystemDistance, min_occurrence=0),
        Element(SystemDividers, min_occurrence=0),
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from unittest import TestCase
from musicscore.musictree.treenote import TreeNote
from musicscore.musicxml.elements.fullnote import Pitch
from musicscore.musicxml.elements.note import Duration


class TestTreeNote(TestCase):
    def setUp(self):
        self.note = TreeNote()

    def test_quarter_duration(self):
        self.note.add_child(Duration(int(self.note.quarter_duration)))
        result = ['Rest', 'Duration']
        self.assertEqual([type(child).__name__ for child in self.note.get_children()], result)
        self.note.quarter_duration = 0
        # result = ['Grace', 'Rest']
        # self.assertEqual([type(child).__name__ for child in self.note.get_children()], result)

    def test_duration(self):
        self.note.quarter_duration = 2
        self.note.update_duration(4)

        self.assertEqual(self.note.duration.value, 8)

    def test_accidental(self):
        self.note.accidental.show = False
        self.note.add_child(Duration(int(self.note.quarter_duration)))
        self.note.event = Pitch(step='A', alter=-1.5)
        self.note.accidental.show = True
        result = '''<note>
  <pitch>
    <step>A</step>
    <alter>-1.5</alter>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <accidental>three-quarters-flat</accidental>
</note>
'''
        self.assertEqual(self.note.to_string(), result)





from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeGroupBarlineValue


class ComplexTypeGroupBarline(ComplexType, TypeGroupBarlineValue, Color):
    """The group-barline type indicates if the group should have common barlines."""

    def __init__(self, tag, value=None, *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract, String
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.font import Font
from musicscore.musicxml.types.complextypes.complextype import ComplexType


class Smulf(AttributeAbstract):
    def __init__(self, smulf=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('smulf', smulf, 'SmulfLyricsGlyphName')


class ComplexTypeElision(ComplexType, String, Font, Color, Smulf):
    """
    The elision type represents an elision between lyric syllables. The text content specifies the symbol used to
    display the elision. Common values are a no-break space (Unicode 00A0), an underscore (Unicode 005F), or an undertie
    (Unicode 203F). If the text content is empty, the smufl attribute is used to specify the symbol to use. Its value
    is a SMuFL canonical glyph name that starts with lyrics. The SMuFL attribute is ignored if the elision glyph is
    already specified by the text content. If neither text content nor a smufl attribute are present, the elision
    glyph is application-specific.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)



import os
from unittest import TestCase

from quicktions import Fraction

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        self.score.add_measure(TreeMeasure(time=(1, 4)))
        sf = SimpleFormat(durations=[Fraction(4, 6), Fraction(1, 6), Fraction(1, 6)])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)



from unittest import TestCase

from quicktions import Fraction

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treepart import TreePart
from musicscore.musictree.treescoretimewise import TreeScoreTimewise


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        self.score.add_measure(TreeMeasure(time=(5, 4)))
        sf = SimpleFormat(durations=[5])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        self.score.finish()
        output = []
        for measure in self.score.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                for beat in part.get_beats():
                    output.append([ch.quarter_duration for ch in beat.chords])

        result = [[Fraction(3, 1)], [], [], [Fraction(2, 1)], []]
        self.assertEqual(output, result)


import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treeclef import ALTO_CLEF
from musicscore.musictree.treeinstruments import TreeInstrument, Violin, Cello, Viola
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def test_1(self):
        instrument = TreeInstrument(name='banjo', abbreviation='bjo', number=2)
        instrument.standard_clef = ALTO_CLEF
        sf = SimpleFormat(durations=[1, 2, 3, 1, 2, 3, 1, 2, 3])
        score = TreeScoreTimewise()
        sf.to_stream_voice(2).add_to_score(score)
        sf.transpose(5)
        sf.to_stream_voice(1).add_to_score(score)
        score.get_score_parts()[0].instrument = instrument

        xml_path = path + '_test_1.xml'
        score.write(xml_path)
        TestScore().assert_template(result_path=xml_path)

    def test_2(self):
        sf = SimpleFormat(durations=[1, 2, 3, 1, 2, 3, 1, 2, 3], midis=9 * [60 - 5])
        score = TreeScoreTimewise()
        for part_number in range(1, 5):
            sf.to_stream_voice().add_to_score(score, part_number=part_number)

        score_parts = score.get_score_parts()
        score_parts[0].instrument = Violin(1)
        score_parts[1].instrument = Violin(1)
        score_parts[2].instrument = Viola()
        score_parts[3].instrument = Cello()
        xml_path = path + '_test_2.xml'
        score.write(xml_path)
        TestScore().assert_template(result_path=xml_path)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.midi import G, C
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treeclef import ALL_CLEFS, SUPER_HIGH_TREBLE_CLEF, SUPER_LOW_BASS_CLEF
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat()
        for clef in ALL_CLEFS:
            if clef not in [SUPER_HIGH_TREBLE_CLEF, SUPER_LOW_BASS_CLEF]:
                for midi in clef.optimal_range:
                    if not midi:
                        if clef.optimal_range.index(midi) == 0:
                            midi = G(0)
                        else:
                            midi = C(8)
                    sf.add_chord(TreeChord(quarter_duration=2, midis=[midi]))
                sf.chords[-2].add_clef(clef)
        xml = path + '_test_1.xml'
        sf.to_stream_voice().add_to_score(self.score)
        self.score.finish()
        self.score.to_partwise()
        self.score.write(xml)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treeinstruments import Violin, Cello, Viola
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()
        sf = SimpleFormat(durations=[4])
        sf.to_stream_voice().add_to_score(self.score, part_number=1)
        sf.to_stream_voice().add_to_score(self.score, part_number=2)
        sf.to_stream_voice().add_to_score(self.score, part_number=3)
        sf.to_stream_voice().add_to_score(self.score, part_number=4)
        self.score.get_score_parts()[0].instrument = Violin(1)
        self.score.get_score_parts()[1].instrument = Violin(2)
        self.score.get_score_parts()[2].instrument = Viola()
        self.score.get_score_parts()[3].instrument = Cello()

    def test_1(self):
        xml_path = path + '_test_1.xml'
        self.score.get_score_parts()[0].add_part_group(number=1, type='start', name='', symbol='bracket', barline='yes')
        self.score.get_score_parts()[2].add_part_group(number=1, type='stop')
        self.score.write(path=xml_path)
        TestScore().assert_template(result_path=xml_path)

    def test_2(self):
        xml_path = path + '_test_2.xml'
        self.score.get_score_parts()[1].add_part_group(number=1, type='start', name='', symbol='bracket', barline='yes')
        self.score.get_score_parts()[2].add_part_group(number=1, type='stop')
        self.score.write(path=xml_path)
        TestScore().assert_template(result_path=xml_path)

    def test_3(self):
        xml_path = path + '_test_3.xml'
        self.score.get_score_parts()[0].add_part_group(number=1, type='start', name='', symbol='bracket', barline='yes')
        self.score.get_score_parts()[3].add_part_group(number=1, type='stop')
        self.score.write(path=xml_path)
        TestScore().assert_template(result_path=xml_path)



from unittest import TestCase

from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musicxml.types.complextypes.attributes import Beats


class TestCopyMeasure(TestCase):
    def setUp(self):
        self.measure = TreeMeasure()

    def test_copy_measure(self):
        new_measure = self.measure.__copy__()
        new_measure.time.get_children_by_type(Beats)[0].value = 8

        result = '''<time>
  <beats>8</beats>
  <beat-type>4</beat-type>
</time>
'''
        self.assertEqual(new_measure.time.to_string(), result)

        result = '''<time>
  <beats>4</beats>
  <beat-type>4</beat-type>
</time>
'''
        self.assertEqual(self.measure.time.to_string(), result)




from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechordflags import PercussionFlag, XFlag
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        xml_path = path + '_test_1.xml'
        durations = [2, 2]
        sf = SimpleFormat(durations=durations)
        chords = sf.chords
        chords[0].add_tie('start')
        chords[1].add_tie('stop')
        chords[1].is_adjoinable = False
        # sf.to_stream_voice().add_to_score(self.score, part_number=1)
        sf._chords = []
        new_chords = chords[0].split(1, 1)

        new_chords.append(chords[1])
        for ch in new_chords:
            # ch.is_adjoinable = False
            sf.add_chord(ch)
        v = sf.to_stream_voice()
        # print([ch.quarter_duration for ch in v.chords])
        # print([ch.is_tied_to_next for ch in v.chords])
        # print([ch.is_adjoinable for ch in v.chords])

        sf.to_stream_voice().add_to_score(self.score, part_number=1)

        self.score.write(xml_path)
        # TestScore().assert_template(xml_path)

    def test_2(self):
        xml_path = path + '_test_2.xml'
        durations = [5]
        sf = SimpleFormat(durations=durations)

        chords = sf.chords
        sf._chords = []
        new_chords = chords[0].split(4, 1)
        new_chords[0].is_adjoinable = False
        for ch in new_chords:
            sf.add_chord(ch)

        chords = sf.chords
        sf._chords = []
        new_chords = chords[0].split(1, 3)
        new_chords[0].is_adjoinable = False
        new_chords.append(chords[-1])
        for ch in new_chords:
            sf.add_chord(ch)

        chords = sf.chords
        sf._chords = []
        sf.add_chord(chords[0])
        new_chords = chords[1].split(1, 2)
        new_chords[0].is_adjoinable = False
        new_chords.append(chords[-1])
        for ch in new_chords:
            sf.add_chord(ch)

        chords = sf.chords
        sf._chords = []
        sf.add_chord(chords[0])
        sf.add_chord(chords[1])
        new_chords = chords[2].split(1, 1)
        new_chords[0].is_adjoinable = False
        new_chords.append(chords[-1])
        for ch in new_chords:
            sf.add_chord(ch)

        sf.to_stream_voice().add_to_score(self.score, part_number=1)

        self.score.write(xml_path)
        # TestScore().assert_template(xml_path)



from musicscore.musicxml.elements.note import Note, Grace, Duration, Beam, Tie, Type, Lyric, Accidental
from musicscore.musicxml.elements.fullnote import Chord, Pitch, DisplayStep, DisplayOctave, Rest, Step, Octave, Alter

from unittest import TestCase


class TestAccidentalDTD(TestCase):
    def setUp(self):
        self.note = Note()
        self.note.add_child(Duration())
        pitch = self.note.add_child(Pitch())
        pitch.add_child(Alter(1))
        self.note.add_child(Accidental('sharp'))

    def test_accidental(self):
        result = '''<note>
  <pitch>
    <step>C</step>
    <alter>1</alter>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <accidental>sharp</accidental>
</note>
'''
        self.assertEqual(self.note.to_string(), result)



from musicscore.musicxml.elements.note import Note, Grace, Duration, Beam, Tie, Type, Lyric
from musicscore.musicxml.elements.fullnote import Chord, Pitch, DisplayStep, DisplayOctave, Rest
from musicscore.dtd.dtd import ChildOccurrenceDTDConflict, ChildTypeDTDConflict, ChildIsNotOptional
from unittest import TestCase


class TestNoteBeamDTD(TestCase):
    def setUp(self):
        self.note = Note()
        self.note.add_child(Pitch())
        self.note.add_child(Duration())

    def test_note_beam(self):
        self.note.add_child(Beam('begin', number=2))
        result = '''<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <beam number="2">begin</beam>
</note>
'''
        self.assertEqual(self.note.to_string(), result)


from musicscore.dtd.dtd import Sequence, GroupReference, Element
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.printattributes import PrintAttributes
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.groups.layout import Layout
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.scorepart import PartNameDisplay, PartAbbreviationDisplay


class MeasureLayout(XMLElement):
    _TAG = 'measure-layout'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        raise NotImplementedError()


class MeasureNumbering(XMLElement):
    _TAG = 'measure-numbering'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        raise NotImplementedError()


class ComplexTypePrint(ComplexType, PrintAttributes, OptionalUniqueId):
    """
    The print type contains general printing parameters, including the layout elements defined in the layout.mod file.
    The part-name-display and part-abbreviation-display elements used in the score.mod file may also be used here to
    change how a part name or abbreviation is displayed over the course of a piece. They take effect when the current
    measure or a succeeding measure starts a new system.

    Layout elements in a print statement only apply to the current page, system, staff, or measure. Music that follows
    continues to take the default values from the layout included in the defaults element.
    """
    _DTD = Sequence(
        GroupReference(Layout),
        Element(MeasureLayout,
                min_occurrence=0),
        Element(MeasureNumbering,
                min_occurrence=0),
        Element(PartNameDisplay,
                min_occurrence=0),
        Element(PartAbbreviationDisplay,
                min_occurrence=0)
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.midi import C, D, E, F, B
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        midis = [C(4, 'flat'), C(4), C(4, 'sharp')]
        sf = SimpleFormat(midis=midis)
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_1'
        self.score.write(path=result_path)
        # TestScore().assert_template(result_path=result_path)

    def test_2(self):
        midis = [D(4), D(4), D(4), D(4), C(4), C(4), C(4), B(3), C(4), C(4), C(4), F(4), E(4), D(4), F(4), F(4), F(4),
                 E(4), D(4), E(4), F(4)]

        durations = [2, 3, 1, 2, 3, 1, 4, 2, 4, 2, 2, 3, 1, 2, 2, 2, 2, 3, 1, 1, 4]
        durations = [d / 2 for d in durations]
        sf = SimpleFormat(midis=midis, durations=durations)
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_2'
        self.score.write(path=result_path)

    def test_3(self):
        midi = E(6)
        # print(midi.value)
        midi.transpose(3)
        # print(midi.value)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Dir(AttributeAbstract):
    """"""

    def __init__(self, dir=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('dir', dir, "TextDirection")


class TextDirection(Dir):
    """
    The text-direction attribute is used to adjust and override the Unicode bidirectional text algorithm, similar to
    the W3C Internationalization Tag Set recommendation.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


"""
	<xs:attributeGroup name="text-direction">
		<xs:annotation>
			<xs:documentation></xs:documentation>
		</xs:annotation>
		<xs:attribute name="dir" type="text-direction"/>
	</xs:attributeGroup>
"""



from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeCaesuraValue


# class ComplexTypeCaesura(ComplexType, TypeCaesuraValue, PrintStyle, Placement):
#     """The caesura element indicates a slight pause. It is notated using a "railroad tracks" symbol or other variations
#     specified in the element content."""
#
#     def __init__(self, tag, value, *args, **kwargs):
#         super().__init__(tag=tag, value=value, *args, **kwargs)

class ComplexTypeCaesura(ComplexType, PrintStyle, Placement):
    """The caesura element indicates a slight pause. It is notated using a "railroad tracks" symbol or other variations
    specified in the element content."""

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.elements.barline import Barline
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.dtd.dtd import Sequence, Choice, Element
from musicscore.musicxml.elements.note import Note
from musicscore.musicxml.types.complextypes.attributes import ComplexTypeAttributes
from musicscore.musicxml.types.complextypes.backup import ComplexTypeBackup
from musicscore.musicxml.types.complextypes.direction import ComplexTypeDirection
from musicscore.musicxml.types.complextypes.sound import ComplexTypeSound
from musicscore.musicxml.types.complextypes.print import ComplexTypePrint


class Backup(ComplexTypeBackup):
    _TAG = 'backup'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Forward(XMLElement):
    _TAG = 'forward'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        NotImplementedError()


class Direction(ComplexTypeDirection):
    _TAG = 'direction'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Attributes(ComplexTypeAttributes):
    _TAG = 'attributes'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Print(ComplexTypePrint):
    _TAG = 'print'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Sound(ComplexTypeSound):
    _TAG = 'sound'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Link(XMLElement):
    _TAG = 'link'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        NotImplementedError()


class Bookmark(XMLElement):
    _TAG = 'bookmark'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        NotImplementedError()


MusicData = Sequence(
    Choice(
        Element(Note),
        Element(Backup),
        Element(Forward),
        Element(Direction),
        Element(Attributes),
        Element(Print),
        Element(Sound),
        Element(Barline),
        Element(Link),
        Element(Bookmark),
        min_occurrence=0,
        max_occurrence=None
    )
)

'''
                <xs:element name="note" type="note"/>
                <xs:element name="backup" type="backup"/>
                <xs:element name="forward" type="forward"/>
                <xs:element name="direction" type="direction"/>
                <xs:element name="attributes" type="attributes"/>
                <xs:element name="harmony" type="harmony"/>
                <xs:element name="figured-bass" type="figured-bass"/>
                <xs:element name="print" type="print"/>
                <xs:element name="sound" type="sound"/>
                <xs:element name="barline" type="barline"/>
                <xs:element name="grouping" type="grouping"/>
                <xs:element name="link" type="link"/>
                <xs:element name="bookmark" type="bookmark"/>
'''



from unittest import TestCase
from musicscore.musicxml.elements.note import Duration


class TestDuration(TestCase):
    def setUp(self):
        self.duration = Duration(10)

    def test_duration(self):
        result = '''<duration>10</duration>
'''
        self.assertEqual(self.duration.to_string(), result)
        self.duration.value = 2
        result = '''<duration>2</duration>
'''
        self.duration.text = 3
        result = '''<duration>3</duration>
'''
        self.assertEqual(self.duration.to_string(), result)



"""
    <xs:attributeGroup name="printout">
        <xs:annotation>
            <xs:documentation>
            </xs:documentation>
        </xs:annotation>
        <xs:attributeGroup ref="print-object"/>
        <xs:attribute name="print-dot" type="yes-no"/>
        <xs:attributeGroup ref="print-spacing"/>
        <xs:attribute name="print-lyric" type="yes-no"/>
    </xs:attributeGroup>
"""

from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.attributes.printobject import PrintObject
from musicscore.musicxml.attributes.printspacing import PrintSpacing


class PrintDot(AttributeAbstract):
    """"""

    def __init__(self, print_dot=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('print-dot', print_dot, 'TypeYesNo')


class PrintLyric(AttributeAbstract):
    """"""

    def __init__(self, print_lyric=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('print-lyric', print_lyric, 'TypeYesNo')


class Printout(PrintObject, PrintDot, PrintSpacing, PrintLyric):
    """The printout attribute group collects the different controls over printing an object (e.g. a note or rest) and
    its parts, including augmentation dots and lyrics. This is especially useful for notes that overlap in different
    voices, or for chord sheets that contain lyrics and chords but no melody.

    By default, all these attributes are set to yes. If print-object is set to no, the print-dot and print-lyric
    attributes are interpreted to also be set to no if they are not present."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from lxml import etree as et

from musicscore.basic_functions import replace_dash
from musicscore.dtd.dtd import DTDError, ChildIsNotOptional, ChildTypeDTDConflict, ChildOccurrenceDTDConflict, Sequence
from musicscore.musicxml.exceptions import AfterInitializationError
from musicscore.tree.tree import Tree


class XMLTree(Tree):
    _DTD = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dtd = self._DTD
        self._current_children = []
        self._current_choice = None
        self._choice_index = 0
        self._old_choice_index = None
        self._not_sorted_children = []

    @property
    def current_children(self):
        return self._current_children

    def get_choice_index(self):
        return self._choice_index

    def _set_child(self, type_, tag, value):
        if value is not None and not isinstance(value, type_):
            value = type_(value)

        current_xml_children = self.current_children
        if value:
            for xml_child in current_xml_children:
                if isinstance(xml_child, type_):
                    self.current_dtd_choice.remove_xml_child(xml_child)

            self.add_xml_child(value)

        name = '_' + replace_dash(tag)

        self.__setattr__(name, value)

    @property
    def current_dtd_choice(self):
        if self._old_choice_index != self._choice_index:
            self._current_choice = self.dtd.get_choices()[self._choice_index].carbon_copy()
            self._old_choice_index = self._choice_index
        return self._current_choice

    def goto_next_dtd_choice(self):
        try:
            old_xml_children = self._not_sorted_children
            self._not_sorted_children = []
            self._choice_index += 1

            for old_xml_child in old_xml_children:
                self.add_xml_child(old_xml_child)
        except IndexError:
            raise StopIteration()

    def check_non_optional(self):
        for leaf in self.current_dtd_choice.traverse_leaves():
            if not leaf.check_min_occurrence():
                raise ChildIsNotOptional(leaf, self)

    def reset_dtd(self):
        self._not_sorted_children = []
        self._current_choice = None
        self._choice_index = 0
        self._old_choice_index = None

    def add_xml_child(self, xml_child):
        choice = self.current_dtd_choice
        selected_leaves = [leaf for leaf in choice.traverse_leaves() if isinstance(xml_child, leaf.type_)]
        # print(selected_leaves)
        if not selected_leaves:
            try:
                self.goto_next_dtd_choice()
                self.add_xml_child(xml_child)
            except Exception:
                raise ChildTypeDTDConflict(xml_child, self)

        else:
            child_added = False
            for selected_leaf in selected_leaves:
                parent = selected_leaf.up
                child_added = selected_leaf.add_xml_child(xml_child)
                if child_added:
                    break

            if not child_added and isinstance(parent, Sequence) and (parent.min_occurrence, parent.max_occurrence) != (
                    1, 1):
                try:
                    parent.pattern
                except AttributeError:
                    parent.pattern = [child.__deepcopy__() for child in parent.get_children()]
                for child in parent.pattern:
                    parent.add_child(child.__deepcopy__())

                child_added = self.add_xml_child(xml_child)

            if not child_added:
                try:
                    self.goto_next_dtd_choice()
                    self.add_xml_child(xml_child)
                except Exception as err:
                    raise ChildOccurrenceDTDConflict(xml_child, self)

            if child_added:
                self.update_current_children()
                self._not_sorted_children.append(xml_child)

            return child_added

    def update_current_children(self):
        self._current_children = self.current_dtd_choice.get_xml_children()

    def add_child(self, child):
        if not self.dtd:
            raise DTDError('_DTD is None. No Child can be added. ')
        self.add_xml_child(child)
        child._up = self
        return child

    def remove_child(self, child):

        current_xml_children = self.current_children
        if child:
            current_xml_children.remove(child)

        self.reset_dtd()

        for xml_child in current_xml_children:
            self.add_xml_child(xml_child)

    def close_dtd(self):
        if self.dtd:
            try:
                self.check_non_optional()
            except ChildIsNotOptional as e:
                try:
                    self.goto_next_dtd_choice()
                    self.close_dtd()
                except (DTDError, StopIteration):
                    raise e
        for child in self.get_children():
            child.close_dtd()

    def get_children(self):
        return self.current_children

    def get_children_by_type(self, type_):
        return [child for child in self.current_children if isinstance(child, type_)]


class XMLElement(XMLTree):
    _ATTRIBUTES = None

    def __init__(self, tag, *args, **kwargs):
        self._attributes = {}
        self._tag = None
        self.tag = tag
        super().__init__(*args, **kwargs)
        self._text = None

    @property
    def tag(self):
        return self._tag

    @tag.setter
    def tag(self, value):
        if self._tag is not None:
            raise AfterInitializationError(self.tag)
        self._tag = value

    @property
    def text(self):
        try:
            return self.value
        except AttributeError:
            return self._text

    @text.setter
    def text(self, v):
        try:
            self.value = v
        except AttributeError:
            self._text = v

    def __getattr__(self, item):
        tag = replace_dash(item)
        found_children = self.get_children_by_tag(tag)
        if len(found_children) == 0:
            raise AttributeError('object "{}" has no attribute "{}"'.format(type(self).__name__, item))

        if len(found_children) == 1:
            return found_children[0]
        else:
            return found_children

    def get_attributes(self):
        return self._attributes

    def get_attribute(self, attribute_name):
        return self.get_attributes()[attribute_name]

    def get_children_by_tag(self, tag):
        return [child for child in self.get_children() if child.tag == tag]

    def _sort_attributes(self):
        def sort_function(x):
            return self._ATTRIBUTES.index(x)

        sorted_attributes = {}
        for key in sorted(self._attributes, key=sort_function):
            sorted_attributes[key] = self._attributes[key]
        self._attributes = sorted_attributes

    def set_attribute(self, attribute_name, attribute_value):
        if self._ATTRIBUTES is not None:
            if attribute_name not in self._ATTRIBUTES:
                raise ValueError('{}.set_attribute: attribute_name: {} is not in {}._ATTRIBUTES'.format(type(self),
                                                                                                        attribute_name,
                                                                                                        self.__class__.__name__))
        self.get_attributes()[attribute_name] = attribute_value
        if self._ATTRIBUTES is not None:
            self._sort_attributes()

    def remove_attribute(self, attribute):
        if attribute in self.get_attributes().keys():
            self.get_attributes().pop(attribute)

    def clear_attributes(self):
        self.get_attributes().clear()

    def __repr__(self):
        try:
            return '{} {} at {}'.format(self.__class__.__name__, self.value, hex(id(self)))
        except AttributeError:
            return '{} at {}'.format(self.__class__.__name__, hex(id(self)))

    def _to_xml(self):
        xml = et.Element(_tag=self.tag)

        def set_attributes():
            for key in self.get_attributes().keys():
                xml.set(key, str(self.get_attributes()[key]))

        def set_children():
            for child in self.get_children():
                if isinstance(child, XMLElement):
                    xml.append(child._to_xml())
                else:
                    raise TypeError('child {} must be of type XMLElement'.format(child))

        set_children()
        if self.text is not None:
            xml.text = str(self.text)
        set_attributes()
        return xml

    def to_string(self):
        self.close_dtd()
        xml = self._to_xml()
        return et.tounicode(xml, pretty_print=True)



from musicscore.musicxml.elements.note import Notehead, TimeModification, Stem, Notations
from musicscore.musicxml.groups.common import Voice
from musicscore.musicxml.types.complextypes.timemodification import ActualNotes, NormalNotes


class TreeChordFlag(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _get_split(self, beat, chord, minimum_duration):
        if minimum_duration == 1:
            spl = {1.5: [2, 1], 2: [1, 1], 3: [1, 2], 4: [1, 3], 6: [1, 5]}
        elif minimum_duration == 0.5:
            spl = {1: [1, 1], 1.5: [1, 2], 2: [1, 3], 3: [1, 5], 4: [1, 7], 6: [1, 11]}
        else:
            raise ValueError('minimum_duration can only be 1 or 0.5')

        try:
            return chord.split(*spl[chord.quarter_duration])
        except KeyError:
            return [chord]

    def implement(self, chord, beat):
        raise NotImplementedError('TreeChordFlag.implement() must be overwritten')

        # if beat.duration == 1:
        #     try:
        #         return chord.split(*spl[chord.quarter_duration])
        #     except KeyError:
        #         return [chord]

    def implement_percussion_notation(self, chord, beat, minimum_duration=1):
        if chord.is_tied_to_next:
            chord.remove_tie('start')
        if chord.is_tied_to_previous:
            chord.to_rest()
            output = [chord]
        elif chord.position_in_beat == 0 and chord.is_rest is False:
            output = self._get_split(beat, chord, minimum_duration)
            try:
                output[1].to_rest()
            except IndexError:
                pass
        else:
            output = [chord]
        for chord in output:
            chord.remove_flag(self)
        return output

    def __deepcopy__(self, memodict={}):
        return self.__class__()


class PizzFlag(TreeChordFlag):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def implement(self, chord, beat):
        output = self.implement_percussion_notation(chord, beat)
        # for ch in output:
        #     if not ch.is_rest:
        #         ch.add_words('pizz.')
        return output


class PercussionFlag(TreeChordFlag):
    def __init__(self, minimum_duration=1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.minimum_duration = minimum_duration

    def implement(self, chord, beat):
        output = self.implement_percussion_notation(chord, beat, self.minimum_duration)
        return output

    def __deepcopy__(self, memodict={}):
        return self.__class__(minimum_duration=self.minimum_duration)


class BeatwiseFlag(TreeChordFlag):
    def __init__(self, slur='dashed', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.slur = slur

    def __deepcopy__(self, memodict={}):
        return self.__class__(slur=self.slur)

    def _get_split(self, chord, beat):
        quarter_beat = {2: [1, 1], 3: [1, 2], 4: [1, 3], 6: [1, 5]}
        eighth_beat = {1: [1, 1], 1.5: [1, 2], 2: [1, 3], 3: [1, 5]}

        if beat.duration == 1:
            try:
                return chord.split(*quarter_beat[chord.quarter_duration])
            except KeyError:
                return [chord]

        elif beat.duration == 0.5:

            try:
                split = chord.split(*eighth_beat[chord.quarter_duration])
                # for ch in split:
                #     ch.force_tie()
                return split
            except KeyError:
                return [chord]

    def _substitute_ties(self, chord):
        if chord.is_tied_to_previous:
            if self.slur == 'tie':
                chord.is_adjoinable = False
            else:
                chord.remove_tie('stop')
                if self.slur is not None:
                    chord.add_slur('stop')

        if chord.is_tied_to_next:
            if self.slur == 'tie':
                chord.is_adjoinable = False
            else:
                chord.remove_tie('start')
                if self.slur is not None:
                    chord.add_slur('start', line_type=self.slur)

    def implement(self, chord, beat):
        output = self._get_split(chord, beat)
        self._substitute_ties(output[0])
        return output


class XFlag(BeatwiseFlag):
    def __init__(self, slur='dashed', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.slur = slur

    def __deepcopy__(self, memodict={}):
        return self.__class__(slur=self.slur)

    def implement(self, chord, beat):
        output = super().implement(chord, beat)
        output[0].add_child(Notehead('x'))
        return output


class FingerTremoloFlag(BeatwiseFlag):
    def __init__(self, tremolo_chord, number=3, mode='conventional', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mode = None
        self.mode = mode
        self._tremolo_chord = None
        self.tremolo_chord = tremolo_chord
        self.number = number
        self.slur = None

    def __deepcopy__(self, memodict={}):
        return self.__class__(tremolo_chord=self.tremolo_chord.tremolo_flag_copy(), number=self.number, mode=self.mode)

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, val):
        permitted = ['conventional', 'modern']
        if val not in permitted:
            raise ValueError('mode.value {} must be of in {}'.format(type(val), permitted))
        self._mode = val

    @property
    def tremolo_chord(self):
        return self._tremolo_chord

    @tremolo_chord.setter
    def tremolo_chord(self, val):
        if val is not None and self.mode == 'modern':
            val.is_finger_tremolo = True
            val.add_child(Stem('none'))
            val.set_manual_type('quarter', size='full')
            tm = TimeModification()
            tm.add_child(ActualNotes(0))
            tm.add_child(NormalNotes(1))
            val.add_child(tm)
            val.is_adjoinable = False
            val.quarter_duration = 0

        self._tremolo_chord = val

    def _implement_conventional(self, chord, beat):
        self.slur = None
        output = super().implement(chord, beat)
        # output = [chord]

        output[0].quarter_duration /= 2
        output[0].add_tremolo(type='start')

        if not self.tremolo_chord.tremoli:
            self.tremolo_chord.add_tremolo(type='stop')

        self.tremolo_chord.quarter_duration = chord.quarter_duration
        self.tremolo_chord.parent_voice = chord.parent_voice
        self.tremolo_chord.parent_beat = chord.parent_beat
        v = chord.get_children_by_type(Voice)[0]

        if not self.tremolo_chord.get_children_by_type(Voice):
            self.tremolo_chord.add_child(v)

        output.insert(1, self.tremolo_chord)
        return output

    def _implement_modern(self, chord, beat):

        if self.tremolo_chord.midis[0].value < chord.midis[0].value:
            chord.add_words('\uF415', font_family='bravura', font_size=16, relative_x=30, relative_y=-50)
            chord.set_tie_orientation('over')
        else:
            chord.add_words('\uF417', font_family='bravura', font_size=16, relative_x=30, relative_y=-50)
            chord.set_tie_orientation('under')
        self.tremolo_chord.parent_voice = chord.parent_voice
        self.tremolo_chord.parent_beat = chord.parent_beat
        v = chord.get_children_by_type(Voice)[0]
        if not self.tremolo_chord.get_children_by_type(Voice):
            self.tremolo_chord.add_child(v)

        return [chord, self.tremolo_chord]

    def implement(self, chord, beat):
        if self.mode == 'modern':
            return self._implement_modern(chord, beat)
        elif self.mode == 'conventional':
            return self._implement_conventional(chord, beat)


class GlissFlag(BeatwiseFlag):
    def __init__(self, mode=1, clef='bass', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._head = True
        self.slur = None
        self.clef = clef
        self.mode = mode

    def __deepcopy__(self, memodict={}):
        copied = self.__class__(mode=self.mode, clef=self.clef)
        copied._head = False
        return copied

    def implement(self, chord, beat):
        output = super().implement(chord, beat)
        if not self._head:
            if output[0].quarter_duration == 1:
                if self.clef == 'bass':
                    midi = 59
                elif self.clef == 'treble':
                    midi = 79
                else:
                    raise NotImplementedError()
                output[0].midis = [midi]
            output[0].add_child(Notehead('none'))

        else:
            if self.mode != 1:
                output[0].add_slide('stop')
            output[0].add_slide('start')
        return output



import math
import warnings

from quicktions import Fraction

from musicscore.basic_functions import flatten
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treechordflags import TreeChordFlag, FingerTremoloFlag
from musicscore.musicxml.elements.note import TimeModification
from musicscore.musicxml.types.complextypes.timemodification import ActualNotes, NormalNotes, NormalType


class BeatException(BaseException):
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)


class WrongPositionInBeat(BeatException):
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)


def _find_nearest_quantized_value(quantized_values, values):
    output = []
    for value in values:
        nearest_quantized = min(enumerate(quantized_values), key=lambda x: abs(x[1] - value))[1]
        delta = nearest_quantized - value
        output.append((nearest_quantized, delta))
    return output


def _find_q_delta(quantized_locations, values):
    qs = _find_nearest_quantized_value(quantized_locations, values)
    d = 0
    for q in qs:
        d += abs(q[1])
    return d


def _find_quantized_locations(duration, subdivision):
    output = range(subdivision + 1)
    fr = duration / subdivision
    output = [x * fr for x in output]
    return output


class TreeBeat(object):
    def __init__(self, duration=1, max_division=None, forbidden_divisions=None):
        self._duration = None
        self._max_division = None
        self._forbidden_divisions = None
        self._best_div = None
        self._permitted_durations = (4, 2, 1, 0.5)
        self._chords = []
        self.parent_voice = None

        self.duration = duration
        self.max_division = max_division
        self.forbidden_divisions = forbidden_divisions

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, value):
        if self._duration is None:
            if value not in self._permitted_durations:
                raise ValueError('beat_duration can only be in {}'.format(self._permitted_durations))
        else:
            raise Exception('duration can be set by initialization')

        self._duration = Fraction(value)

    @property
    def max_division(self):
        if self._max_division is None:
            parent_max_division = self.parent_voice.max_division
            if parent_max_division:
                self._max_division = math.floor(parent_max_division * self.duration)
                if self._max_division == 0:
                    self._max_division = 1

        if self._max_division is None:
            if self.duration == 0.5:
                self._max_division = 4
            else:
                self._max_division = 8

        return self._max_division

    @max_division.setter
    def max_division(self, value):
        if value is not None and not isinstance(value, int):
            raise TypeError('max_division.value must be None or of type int not {}'.format(type(value)))
        self._max_division = value

    @property
    def forbidden_divisions(self):
        if self._forbidden_divisions is None:
            self._forbidden_divisions = self.parent_voice.forbidden_divisions

        if self._forbidden_divisions is None:
            self._forbidden_divisions = []

        return self._forbidden_divisions

    @forbidden_divisions.setter
    def forbidden_divisions(self, value):
        if value is not None:
            for x in value:
                if not isinstance(x, int):
                    raise TypeError('forbidden_division must be of type int  not{}'.format(type(value)))
        # max_division = int(max_division / 2)
        # forbidden_divisions = [div // 2 for div in forbidden_divisions if div // 2 != 0]
        # forbidden_divisions = list(set(forbidden_divisions))
        self._forbidden_divisions = value

    @property
    def best_div(self):
        return self._best_div

    @property
    def chords(self):
        return self._chords

    @property
    def tree_part_voice(self):
        return self.parent_voice

    @property
    def previous(self):
        if not self.tree_part_voice:
            raise Exception('beat has no voice')
        index = self.tree_part_voice.beats.index(self)
        if index == 0:
            return None
        return self.tree_part_voice.beats[index - 1]

    @property
    def next(self):
        if not self.tree_part_voice:
            raise Exception('beat has no voice')
        index = self.tree_part_voice.beats.index(self)
        if index == len(self.tree_part_voice.beats) - 1:
            return None
        return self.tree_part_voice.beats[index + 1]

    @property
    def offset(self):
        if self.previous:
            output = self.previous.offset + self.previous.duration
            return output
        else:
            return 0

    @property
    def end_position(self):
        return self.offset + self.duration

    def add_chord(self, chord):
        if not isinstance(chord, TreeChord):
            raise TypeError('{} must be of type TreeChord'.format(chord))
        chord.parent_beat = self
        self.chords.append(chord)
        if chord.position_in_beat > self.duration:
            self.chords.remove(chord)
            raise WrongPositionInBeat()

    def remove_chords(self):
        self._chords = []

    def clear_chords(self):
        self._chords = []

    def get_quantized_locations(self, subdivision):
        return _find_quantized_locations(self.duration, subdivision)

    def get_quantized_durations(self, durations):
        diff = self.duration - sum(durations)
        if diff != 0:
            for index, dur in enumerate(durations):
                durations[index] = dur + diff / len(durations)
        if sum(durations) != self.duration:
            warnings.warn('TreeBeat.get_quarter_durations: sum of durations {} is not equal to beat duration {}'.format(
                sum(durations), self.duration))

        def _get_positions():
            positions = [0]
            for index, duration in enumerate(durations):
                positions.append(positions[index] + duration)
            return positions

        def _get_permitted_divs():
            output = list(range(1, self.max_division + 1))
            for f in self.forbidden_divisions:
                if f in output:
                    output.remove(f)
            return output

        permitted_divs = _get_permitted_divs()
        positions = _get_positions()

        best_div = permitted_divs.pop(0)
        last_q_delta = _find_q_delta(self.get_quantized_locations(subdivision=best_div), positions)

        for div in permitted_divs:
            current_q_delta = _find_q_delta(self.get_quantized_locations(subdivision=div), positions)

            if current_q_delta < last_q_delta:
                best_div = div
                last_q_delta = current_q_delta

            elif (current_q_delta == last_q_delta) and (div < best_div):
                best_div = div

        quantized_positions = [f[0] for f in
                               _find_nearest_quantized_value(self.get_quantized_locations(subdivision=best_div),
                                                             positions)]
        quantized_durations = []

        for index, duration in enumerate(durations):
            quarter_duration = Fraction(
                quantized_positions[index + 1] - quantized_positions[index]).limit_denominator(
                int(best_div / self.duration))
            quantized_durations.append(quarter_duration)

        self._best_div = best_div
        return quantized_durations

    def quantize(self):
        quarter_durations = [chord.quarter_duration for chord in self.chords]
        if len([d for d in quarter_durations if d != 0]) > 1:
            quarter_durations = self.get_quantized_durations(quarter_durations)
            for chord, quarter_duration in zip(self.chords, quarter_durations):
                chord.quarter_duration = quarter_duration
                chord._offset = None

    def split_not_notatable(self):
        # print(self.offset, [ch.quarter_duration for ch in self.chords])
        output = []
        for chord in self.chords:
            split = None
            if chord.quarter_duration == Fraction(5, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(3, 2)
            elif chord.quarter_duration == Fraction(7, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(4, 3)
            elif chord.quarter_duration == Fraction(9, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(4, 5)
            elif chord.quarter_duration == Fraction(10, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(4, 6)
            elif chord.quarter_duration == Fraction(11, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(4, 7)
            elif chord.quarter_duration == Fraction(13, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 5)
            elif chord.quarter_duration == Fraction(14, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 6)
            elif chord.quarter_duration == Fraction(15, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 7)
            elif chord.quarter_duration == Fraction(16, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 8)
            elif chord.quarter_duration == Fraction(17, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 9)
            elif chord.quarter_duration == Fraction(18, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 10)
            elif chord.quarter_duration == Fraction(19, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 11)
            elif chord.quarter_duration == Fraction(20, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 12)
            elif chord.quarter_duration == Fraction(21, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 13)
            elif chord.quarter_duration == Fraction(22, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 14)
            elif chord.quarter_duration == Fraction(23, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 15)
            elif chord.quarter_duration == Fraction(24, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 16)
            elif chord.quarter_duration == Fraction(25, 1):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 17)
            elif chord.quarter_duration == Fraction(3, 2):
                if chord.position_in_beat == 0.5:
                    split = chord.split(1, 2)
            elif chord.quarter_duration == Fraction(5, 2):
                if chord.position_in_beat == 0:
                    split = chord.split(2, 3)
                elif chord.position_in_beat == 0.5:
                    split = chord.split(1, 4)
            elif chord.quarter_duration == Fraction(7, 2):
                if chord.position_in_beat == 0:
                    split = chord.split(4, 3)
                elif chord.position_in_beat == 0.5:
                    split = chord.split(1, 6)
            elif chord.quarter_duration == Fraction(9, 2):
                if chord.position_in_beat == 0:
                    split = chord.split(8, 1)

            elif chord.quarter_duration == Fraction(1, 4):
                if chord.position_in_beat == Fraction(3, 8):
                    split = chord.split(1, 1)
            elif chord.quarter_duration == Fraction(3, 8):
                if chord.position_in_beat == Fraction(3, 8):
                    split = chord.split(1, 2)
            elif chord.quarter_duration == Fraction(5, 6) and self.best_div == 6:
                if chord.position_in_beat == Fraction(0, 6):
                    split = chord.split(4, 1)
                elif chord.position_in_beat == Fraction(1, 6):
                    split = chord.split(1, 4)

            elif chord.quarter_duration == Fraction(5, 7):
                if chord.position_in_beat == Fraction(0, 7):
                    split = chord.split(3, 2)
                elif chord.position_in_beat == Fraction(1, 7):
                    split = chord.split(3, 2)
                elif chord.position_in_beat == Fraction(2, 7):
                    split = chord.split(2, 3)

            elif chord.quarter_duration == Fraction(5, 8):
                if chord.position_in_beat == Fraction(0, 8):
                    split = chord.split(4, 1)
                elif chord.position_in_beat == Fraction(1, 8):
                    split = chord.split(3, 2)
                elif chord.position_in_beat == Fraction(2, 8):
                    split = chord.split(2, 3)
                elif chord.position_in_beat == Fraction(3, 8):
                    split = chord.split(1, 4)
                else:
                    raise Exception('something wrong with split')


            # elif chord.quarter_duration == Fraction(7,8):
            #     if chord.position_in_beat == Fraction(0,8):
            #         split = chord.split(Fraction(4,8), Fraction(3,8))
            #     elif chord.position_in_beat == Fraction(1,8):
            #         split = chord.split(Fraction(3,8), Fraction(4,8))

            elif chord.quarter_duration == Fraction(5, 9):
                if chord.position_in_beat == Fraction(0, 9):
                    split = chord.split(4, 1)
                elif chord.position_in_beat == Fraction(1, 9):
                    split = chord.split(1, 4)
                elif chord.position_in_beat == Fraction(2, 9):
                    split = chord.split(1, 4)
                elif chord.position_in_beat == Fraction(3, 9):
                    split = chord.split(1, 4)
                elif chord.position_in_beat == Fraction(4, 9):
                    split = chord.split(4, 1)

            elif chord.quarter_duration == Fraction(7, 9):
                if chord.position_in_beat == Fraction(0, 9):
                    split = chord.split(6, 1)
                elif chord.position_in_beat == Fraction(1, 9):
                    split = chord.split(2, 5)
                elif chord.position_in_beat == Fraction(2, 9):
                    split = chord.split(1, 6)

            elif chord.quarter_duration == Fraction(5, 10) and self.best_div == 10:
                if chord.position_in_beat == Fraction(0, 10):
                    split = chord.split(3, 2)
                elif chord.position_in_beat == Fraction(1, 10):
                    split = chord.split(3, 2)
                elif chord.position_in_beat == Fraction(2, 10):
                    split = chord.split(2, 3)
                elif chord.position_in_beat == Fraction(3, 10):
                    split = chord.split(3, 2)
                elif chord.position_in_beat == Fraction(4, 10):
                    split = chord.split(2, 3)
                elif chord.position_in_beat == Fraction(5, 10):
                    split = chord.split(3, 2)

            elif chord.quarter_duration == Fraction(6, 10) and self.best_div == 10:
                if chord.position_in_beat == Fraction(0, 10):
                    split = chord.split(3, 3)
                elif chord.position_in_beat == Fraction(3, 10):
                    split = chord.split(3, 3)
                elif chord.position_in_beat == Fraction(4, 10):
                    split = chord.split(3, 3)

            elif chord.quarter_duration == Fraction(7, 10):
                if chord.position_in_beat == Fraction(0, 10):
                    split = chord.split(3, 4)
                elif chord.position_in_beat == Fraction(1, 10):
                    split = chord.split(3, 4)
                elif chord.position_in_beat == Fraction(2, 10):
                    split = chord.split(1, 6)
                elif chord.position_in_beat == Fraction(3, 10):
                    split = chord.split(4, 3)

            elif chord.quarter_duration == Fraction(9, 10):
                if chord.position_in_beat == Fraction(0, 10):
                    split = chord.split(3, 6)
                elif chord.position_in_beat == Fraction(1, 10):
                    split = chord.split(3, 6)
            # elif chord.quarter_duration == Fraction(5, 11):
            #     if chord.position_in_beat == Fraction(0, 11):
            #         split = chord.split(Fraction(4, 11), Fraction(1, 11))
            #     elif chord.position_in_beat == Fraction(1, 11):
            #         split = chord.split(Fraction(3, 11), Fraction(2, 11))
            #     elif chord.position_in_beat == Fraction(2, 11):
            #         split = chord.split(Fraction(2, 11), Fraction(3, 11))
            #     elif chord.position_in_beat == Fraction(3, 11):
            #         split = chord.split(Fraction(1, 11), Fraction(4, 11))
            #     elif chord.position_in_beat == Fraction(4, 11):
            #         split = chord.split(Fraction(4, 11), Fraction(1, 11))
            #     elif chord.position_in_beat == Fraction(5, 11):
            #         split = chord.split(Fraction(3, 11), Fraction(2, 11))
            #     elif chord.position_in_beat == Fraction(6, 11):
            #         split = chord.split(Fraction(3, 11), Fraction(2, 11))
            #
            # elif chord.quarter_duration == Fraction(7, 11):
            #     if chord.position_in_beat == Fraction(0, 11):
            #         split = chord.split(Fraction(4, 11), Fraction(3, 11))
            #     elif chord.position_in_beat == Fraction(1, 11):
            #         split = chord.split(Fraction(3, 11), Fraction(4, 11))
            #     elif chord.position_in_beat == Fraction(2, 11):
            #         split = chord.split(Fraction(4, 11), Fraction(3, 11))
            #     elif chord.position_in_beat == Fraction(3, 11):
            #         split = chord.split(Fraction(3, 11), Fraction(4, 11))
            #     elif chord.position_in_beat == Fraction(4, 11):
            #         split = chord.split(Fraction(4, 11), Fraction(3, 11))
            #
            # elif chord.quarter_duration == Fraction(10, 11):
            #     if chord.position_in_beat == Fraction(0, 11):
            #         split = chord.split(Fraction(4, 11), Fraction(6, 11))
            #     elif chord.position_in_beat == Fraction(1, 11):
            #         split = chord.split(Fraction(4, 11), Fraction(6, 11))
            #
            # elif chord.quarter_duration == Fraction(9, 11):
            #     if chord.position_in_beat == Fraction(0, 11):
            #         split = chord.split(Fraction(8, 11), Fraction(1, 11))
            #     elif chord.position_in_beat == Fraction(1, 11):
            #         split = chord.split(Fraction(1, 11), Fraction(8, 11))
            #     elif chord.position_in_beat == Fraction(2, 11):
            #         split = chord.split(Fraction(1, 11), Fraction(8, 11))
            #
            # elif chord.quarter_duration == Fraction(5, 12):
            #     if chord.position_in_beat == Fraction(0, 12):
            #         split = chord.split(Fraction(4, 12), Fraction(1, 12))
            #     elif chord.position_in_beat == Fraction(1, 12):
            #         split = chord.split(Fraction(3, 12), Fraction(2, 12))
            #     elif chord.position_in_beat == Fraction(2, 12):
            #         split = chord.split(Fraction(2, 12), Fraction(3, 12))
            #     elif chord.position_in_beat == Fraction(3, 12):
            #         split = chord.split(Fraction(1, 12), Fraction(4, 12))
            #     elif chord.position_in_beat == Fraction(4, 12):
            #         split = chord.split(Fraction(4, 12), Fraction(1, 12))
            #     elif chord.position_in_beat == Fraction(5, 12):
            #         split = chord.split(Fraction(3, 12), Fraction(2, 12))
            #     elif chord.position_in_beat == Fraction(6, 12):
            #         split = chord.split(Fraction(2, 12), Fraction(3, 12))
            #     elif chord.position_in_beat == Fraction(7, 12):
            #         split = chord.split(Fraction(1, 12), Fraction(4, 12))
            #
            # elif chord.quarter_duration == Fraction(6, 12):
            #     if chord.position_in_beat == Fraction(5, 12):
            #         split = chord.split(Fraction(3, 12), Fraction(3, 12))
            #
            # elif chord.quarter_duration == Fraction(7, 12):
            #     if chord.position_in_beat == Fraction(0, 12):
            #         split = chord.split(Fraction(4, 12), Fraction(3, 12))
            #     elif chord.position_in_beat == Fraction(1, 12):
            #         split = chord.split(Fraction(3, 12), Fraction(4, 12))
            #     elif chord.position_in_beat == Fraction(2, 12):
            #         split = chord.split(Fraction(4, 12), Fraction(3, 12))
            #     elif chord.position_in_beat == Fraction(3, 12):
            #         split = chord.split(Fraction(3, 12), Fraction(4, 12))
            #     elif chord.position_in_beat == Fraction(4, 12):
            #         split = chord.split(Fraction(4, 12), Fraction(3, 12))
            #     elif chord.position_in_beat == Fraction(5, 12):
            #         split = chord.split(Fraction(3, 12), Fraction(4, 12))
            #
            # elif chord.quarter_duration == Fraction(9, 12) and self.best_div == 12:
            #     if chord.position_in_beat == Fraction(0, 12):
            #         split = chord.split(Fraction(6, 12), Fraction(3, 12))
            #     if chord.position_in_beat == Fraction(1, 12):
            #         split = chord.split(Fraction(3, 12), Fraction(6, 12))
            #     if chord.position_in_beat == Fraction(2, 12):
            #         split = chord.split(Fraction(3, 12), Fraction(6, 12))
            #     if chord.position_in_beat == Fraction(3, 12):
            #         split = chord.split(Fraction(6, 12), Fraction(3, 12))
            #
            # elif chord.quarter_duration == Fraction(10, 12) and self.best_div == 12:
            #     if chord.position_in_beat == Fraction(0, 12):
            #         split = chord.split(Fraction(6, 12), Fraction(4, 12))
            #     if chord.position_in_beat == Fraction(1, 12):
            #         split = chord.split(Fraction(6, 12), Fraction(4, 12))
            #     if chord.position_in_beat == Fraction(2, 12):
            #         split = chord.split(Fraction(6, 12), Fraction(4, 12))
            #
            # elif chord.quarter_duration == Fraction(11, 12):
            #     if chord.position_in_beat == Fraction(0, 12):
            #         split = chord.split(Fraction(8, 12), Fraction(3, 12))
            #
            #     elif chord.position_in_beat == Fraction(1, 12):
            #         split = chord.split(Fraction(3, 12), Fraction(8, 12))
            if split:
                output.extend(split)
            else:
                output.append(chord)
        self.remove_chords()
        for ch in output:
            try:
                self.add_chord(ch)
            except WrongPositionInBeat:
                number_of_next_beats = (self.chords[-1].position_in_beat + self.chords[
                    -1].quarter_duration) / self.duration
                if number_of_next_beats != int(number_of_next_beats):
                    raise Exception()
                next_beat = self.next
                while number_of_next_beats > 1:
                    next_beat = next_beat.next
                    number_of_next_beats -= 1
                if next_beat.chords:
                    raise Exception()
                next_beat.add_chord(ch)
        # self._chords = output

    def implement_flags(self):
        # print('implement flags for', self.offset)
        # print('chords', self.chords)
        # print([ch.quarter_duration for ch in self.chords])

        def check_implement_output(chords):
            if not isinstance(chords, list):
                raise Exception('output of implement can only be a list of chords')

            # if len(chords) not in [1, 2]:
            #     raise Exception('output of implement can have 1 or 2 chords')

            for ch in chords:
                if not isinstance(ch, TreeChord):
                    raise Exception('output of implement can only be a list of chords')

        flag_types = set([flag.__class__ for flag in flatten([chord.flags for chord in self.chords]) if
                          isinstance(flag, TreeChordFlag)])

        while flag_types:
            flag_type = flag_types.pop()
            output = []
            for chord in self.chords:
                try:
                    chord_flag = [flag for flag in chord.flags if isinstance(flag, flag_type)][0]
                    new_chords = chord_flag.implement(chord, self)
                    check_implement_output(new_chords)
                    diff = sum([ch.quarter_duration for ch in new_chords]) - self.duration
                    if len(new_chords) == 1:
                        output.extend(new_chords)
                    else:
                        if diff > 0 and (isinstance(chord_flag, FingerTremoloFlag) or
                                         chord.position_in_beat + new_chords[0].quarter_duration >= self.duration):
                            next_beat = self.next

                            if not next_beat.chords:
                                next_beat.add_chord(new_chords[-1])
                            else:
                                raise Exception('next_beat is not empty.')

                            output.extend(new_chords[:-1])
                        else:
                            output.extend(new_chords)

                except IndexError:
                    output.append(chord)

            self._chords = output
            self.split_not_notatable()

    #
    # def adjoin_rests(self):
    #     _adjoin = True
    #     if len(self.chords) > 1:
    #         for chord in self.chords:
    #             if not chord.is_rest or not chord.is_adjoinable:
    #                 _adjoin = False
    #                 break
    #
    #         if _adjoin:
    #             first_chord = self.chords[0]
    #             first_chord.quarter_duration = self.duration
    #             for chord in self.chords[1:]:
    #                 chord.parent_voice.chords.remove(chord)
    #
    #             self.remove_chords()
    #             self.add_chord(first_chord)

    # chord_iterator = iter(self.chords)

    #
    #
    # def _adjoin(current_chord, next_chord):
    #
    #     def _chords_are_adjoinable():
    #         condition = current_chord.is_adjoinable and next_chord.is_adjoinable
    #         return condition
    #
    #     def _chords_are_rest():
    #         condition = current_chord.is_rest and next_chord.is_rest
    #         # _print_condition('_chords_are_not_rest', condition)
    #         return condition
    #
    #     # def _chords_have_right_positions():
    #     #     # print 'in _chords_have_right_positions', current_chord.name, next_chord.name
    #     #     condition = current_chord.offset % (
    #     #         1) == 0 and next_chord.offset % 1 == 0
    #     #     # _print_condition('_chords_have_right_positions', condition)
    #     #     return condition
    #
    #     # def _result_is_notatable():
    #     #     condition = current_chord.quarter_duration + next_chord.quarter_duration in notatables
    #     #     # _print_condition('_result_is_notatable', condition)
    #     #     return condition
    #
    #     if _chords_are_adjoinable() and _chords_are_rest():
    #         current_chord.quarter_duration += next_chord.quarter_duration
    #         next_chord.marked = True
    #
    #         try:
    #             next_chord = chord_iterator.__next__()
    #             next_chord = _adjoin(current_chord, next_chord)
    #         except StopIteration:
    #             pass
    #
    #     return next_chord
    #
    # adjoin = True
    #
    # try:
    #     current_chord = chord_iterator.__next__()
    #     next_chord = chord_iterator.__next__()
    # except StopIteration:
    #     adjoin = False
    #
    # while adjoin:
    #     try:
    #         next_chord = _adjoin(current_chord, next_chord)
    #         current_chord = next_chord
    #         next_chord = chord_iterator.__next__()
    #     except StopIteration:
    #         break
    #
    # beat_new_chords = []
    # for chord in self.chords:
    #     try:
    #         if chord.marked:
    #             # chord.parent_voice.chords.remove(chord)
    #             pass
    #         else:
    #             beat_new_chords.append(chord)
    #     except AttributeError:
    #         beat_new_chords.append(chord)
    #
    # if self.chords != beat_new_chords:
    #     print([ch.quarter_duration for ch in self.chords])
    #     print([ch.quarter_duration for ch in beat_new_chords])
    #     self.remove_chords()
    #     for ch in beat_new_chords:
    #         self.add_chord(ch)
    #     self.split_not_notatable()

    def update_tuplets(self):
        tuplet_divisions = [3, 5, 6, 7, 9, 10]
        non_grace_chords = [chord for chord in self.chords if
                            chord.quarter_duration != 0 and not chord.is_finger_tremolo]

        if self.best_div in tuplet_divisions:
            for i in range(len(non_grace_chords)):
                if i == 0:
                    non_grace_chords[0].add_tuplet('start')
                elif i == len(non_grace_chords) - 1:
                    non_grace_chords[-1].add_tuplet('stop')
                else:
                    non_grace_chords[i].add_tuplet('continue')

    def substitute_sextoles(self):
        six_divisions = (
            [Fraction(1, 6), Fraction(1, 6), Fraction(1, 6), Fraction(1, 6), Fraction(1, 6), Fraction(1, 6)],
            [Fraction(1, 6), Fraction(1, 6), Fraction(1, 6), Fraction(1, 3), Fraction(1, 6)],
            [Fraction(1, 6), Fraction(1, 6), Fraction(1, 6), Fraction(1, 2)],
            [Fraction(1, 6), Fraction(1, 3), Fraction(1, 3), Fraction(1, 6)],
            [Fraction(1, 6), Fraction(1, 3), Fraction(1, 6), Fraction(1, 6), Fraction(1, 6)],
            [Fraction(1, 6), Fraction(1, 3), Fraction(1, 2)],
            # [Fraction(1,6), Fraction(1,2), Fraction(1,3)],
            [Fraction(1, 6), Fraction(2, 3), Fraction(1, 6)],

            [Fraction(1, 3), Fraction(1, 6), Fraction(1, 3), Fraction(1, 6)],
            [Fraction(1, 3), Fraction(1, 6), Fraction(1, 2)],
            # [Fraction(1,3), Fraction(1,2), Fraction(1,6)],
            [Fraction(1, 2), Fraction(1, 3), Fraction(1, 6)],
            [Fraction(1, 2), Fraction(1, 6), Fraction(1, 6), Fraction(1, 6)]
        )

        if self.best_div == 6:
            non_grace_chords = [chord for chord in self.chords if chord.quarter_duration != 0]
            chords_quarter_durations = [chord.quarter_duration for chord in non_grace_chords]
            if chords_quarter_durations not in six_divisions:
                for chord in non_grace_chords:
                    tm = chord.get_children_by_type(TimeModification)[0]
                    tm.get_children_by_type(ActualNotes)[0].value = 3
                    tm.get_children_by_type(NormalNotes)[0].value = 2
                    tm.get_children_by_type(NormalType)[0].value = 'eighth'



from unittest import TestCase

from musicscore.musictree.midi import midi_to_frequency, frequency_to_midi


class Test(TestCase):
    def test_1(self):
        midi = 60
        self.assertEqual(midi_to_frequency(midi), 261.6255653005986)

    def test_2(self):
        midi = 69
        self.assertEqual(midi_to_frequency(midi, a4=430), 430)

    def test_3(self):
        frequency = 220
        self.assertEqual(frequency_to_midi(frequency), 57)



from unittest import TestCase

from quicktions import Fraction

from musicscore.musictree.treebeat import TreeBeat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treepart import TreePart
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.elements.note import Lyric
from musicscore.musicxml.types.complextypes.lyric import Text

import os

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):

    def setUp(self):
        m = TreeMeasure(time=(4, 4))
        self.part = TreePart(id='one')
        self.part.max_division = 8
        self.part.forbidden_divisions = []
        m.add_child(self.part)
        # self.part.set_beats()

    # def test_add_chord1(self):
    #     self.part.add_chord(TreeChord((60, 61), quarter_duration=4))
    #     self.part.chord_to_notes()
    #     note_1 = self.part.notes[0]
    #     with self.assertRaises(AttributeError):
    #         print(note_1.chord)
    #     note_2 = self.part.notes[1]
    #     self.assertEqual(type(note_2.chord).__name__, 'Chord')
    #     self.assertEqual(note_1.quarter_duration, note_2.quarter_duration, 4)

    def test_previous_chord(self):
        p = self.part
        p.add_chord(TreeChord((60, 61), quarter_duration=1))
        p.add_chord(TreeChord(62, quarter_duration=2))
        p.add_chord(TreeChord(0, quarter_duration=1))

        self.assertEqual(p.chords[0].previous, None)
        self.assertEqual(id(p.chords[1].previous), id(p.chords[0]))
        self.assertEqual(id(p.chords[2].previous), id(p.chords[1]))

    def test_offset(self):
        p = self.part
        p.add_chord(TreeChord((60, 61), quarter_duration=1))
        p.add_chord(TreeChord(62, quarter_duration=1.75))
        p.add_chord(TreeChord(0, quarter_duration=1.25))
        result = [0, 1, 2.75]
        self.assertEqual([chord.offset for chord in p.chords], result)

    def test_split_beats(self):
        p = self.part
        p.add_chord(TreeChord((60, 61), quarter_duration=1))
        p.add_chord(TreeChord(62, quarter_duration=1.75))
        p.add_chord(TreeChord(0, quarter_duration=1.25))
        p.finish()
        # p._add_chords_to_beats()
        # p._split_chords_beatwise()
        result = [0, Fraction(1, 1), Fraction(2, 1), Fraction(11, 4), Fraction(3, 1)]
        self.assertEqual([chord.offset for chord in p.chords], result)

    def test_split_beats_2(self):
        m = TreeMeasure(time=(3, 4))
        p = TreePart(id='one')
        p.max_division = 8
        p.forbidden_divisions = []
        m.add_child(p)
        # p.set_beats()

        p.add_chord(TreeChord(60, quarter_duration=1.4))
        p.add_chord(TreeChord(60, quarter_duration=1.6))
        # p._add_chords_to_beats()
        # p._split_chords_beatwise()
        p.finish()
        result = [Fraction(1, 1), Fraction(2, 5), Fraction(3, 5), Fraction(1, 1)]
        self.assertEqual([chord.quarter_duration for chord in p.chords], result)

    def test_quantize(self):
        m = TreeMeasure(time=(4, 4))
        p = TreePart(id='one')
        p.max_division = 8
        p.forbidden_divisions = []
        m.add_child(p)
        p.add_chord(TreeChord(60, quarter_duration=1))
        p.add_chord(TreeChord(60, quarter_duration=1.2))
        p.add_chord(TreeChord(60, quarter_duration=0.3))
        p.add_chord(TreeChord(60, quarter_duration=0.2))
        p.add_chord(TreeChord(60, quarter_duration=1.3))
        p.finish()

        result = [Fraction(1, 1), Fraction(1, 1), Fraction(1, 6), Fraction(1, 3), Fraction(1, 6), Fraction(1, 3),
                  Fraction(1, 1)]
        self.assertEqual([chord.quarter_duration for chord in p.chords], result)

    def test_beats(self):
        m = TreeMeasure(time=(3, 8, 2, 4))
        p = TreePart(id='one')
        m.add_child(p)
        tree_part_voice = p.get_voice(1)
        tree_part_voice.set_beats()
        result = [0.5, 0.5, 0.5, 1.0, 1.0]
        self.assertEqual([beat.duration for beat in tree_part_voice.beats], result)
        result = [0, 0.5, 1.0, 1.5, 2.5]
        self.assertEqual([beat.offset for beat in tree_part_voice.beats], result)
        # result = [4, 4, 4, 8, 8]
        # self.assertEqual([beat.max_division for beat in tree_part_voice.beats], result)
        # tree_part_voice.beats[3].max_division = 5
        # result = [4, 4, 4, 5, 8]
        # self.assertEqual([beat.max_division for beat in tree_part_voice.beats], result)
        with self.assertRaises(ValueError):
            tree_part_voice.set_beats([TreeBeat(duration=0.5), TreeBeat(duration=0.5), TreeBeat(duration=0.5)])
        tree_part_voice.set_beats([TreeBeat(duration=1), TreeBeat(duration=0.5), TreeBeat(duration=2)])
        result = [0, 1, 1.5]
        self.assertEqual([beat.offset for beat in tree_part_voice.beats], result)

    def test_split_quantize(self):
        s = TreeScoreTimewise()
        m = TreeMeasure(time=(3, 4))
        s.add_measure(m)
        s.add_part()
        # m.add_child(p)

        chord1 = s.add_chord(1, 1, TreeChord((71, 72), quarter_duration=1.3))
        l1 = Lyric()
        l1.add_child(Text('bla'))
        chord1.add_child(l1)
        s.add_chord(1, 1, TreeChord((60, 63, 65), quarter_duration=0.6))
        s.add_chord(1, 1, TreeChord(60, quarter_duration=1.1))
        s.finish()

        # print(s.to_string())
        s.write(path=path)

        result = '''<part id="one">
  <attributes>
    <divisions>1</divisions>
  </attributes>
  <note>
    <pitch>
      <step>B</step>
      <octave>4</octave>
    </pitch>
    <tie/>
    <lyric number="1">
      <text>bla</text>
    </lyric>
  </note>
  <note>
    <pitch>
      <step>B</step>
      <octave>4</octave>
    </pitch>
    <tie/>
  </note>
  <note>
    <pitch>
      <step>B</step>
      <octave>4</octave>
    </pitch>
  </note>
  <note>
    <rest/>
  </note>
</part>
'''
        # self.assertEqual(p.to_string(), result)



from musicscore.dtd.dtd import Element, Sequence, Choice, GroupReference
from musicscore.musicxml.groups.common import Editorial
from musicscore.musicxml.elements.fullnote import FullNote, Pitch, Unpitched
from musicscore.musicxml.types.complextypes.lyric import Elision, EndLine, EndParagraph, Syllabic, Humming, Laughing, \
    Extend, Text
from unittest import TestCase

from musicscore.musicxml.elements.note import Grace, Play, Lyric, Notations, Beam, Staff, NoteheadText, Notehead, Stem, \
    TimeModification, Accidental, Dot, Type, EditorialVoice, Instrument, Tie, DurationGroup, Cue

el = Element(Elision)

seq1 = Sequence(
    Element(Syllabic),
    Element(Elision)
)

seq2 = Sequence(
    Element(Syllabic),
    Element(Elision),
    Sequence(
        Element(EndLine),
        Element(EndParagraph)
    )
)

ch1 = Choice(
    Element(Syllabic),
    Element(Elision),
    Element(EndLine)
)

ch2 = Sequence(
    Choice(
        Element(Syllabic),
        Element(Elision)
    ),
    Element(EndLine)
)

ch3 = Choice(
    Sequence(
        Choice(
            Sequence(
                Element(Syllabic),
                Element(Elision)
            ),
            Element(EndParagraph)
        ),
        Element(EndLine)
    ),
    Element(EndLine)
)

ch4 = Choice(
    Sequence(
        Choice(
            Element(Grace)
        )
    )
)

ch5 = Sequence(
    Choice(
        Sequence(
            Choice(
                Element(Grace)
            )
        )
    )
)

seq3 = Sequence(
    Choice(
        Sequence(
            Choice(
                Element(Pitch),
                Element(Unpitched),
            )

        )
    )
)

# seq4 = GroupReference(FullNote)

# seq4 = Sequence(
#     Element(Chord, min_occurrence=0),
#     Choice(
#         Element(Pitch),
#         Element(Unpitched),
#         Element(Rest)
#     )
# )

seq4 = Sequence(
    GroupReference(FullNote)
)

seq5 = Choice(
    Sequence(
        GroupReference(FullNote),
        Element(Tie, 0, 2)
    )
)

seq6 = Sequence(
    Element(Grace),
    Choice(
        Sequence(
            GroupReference(FullNote)
        ),
        Sequence(
            GroupReference(FullNote),
            Element(Tie, 0, 2)
        ),
        Sequence(
            Element(Cue),
            GroupReference(FullNote)
        )
    )
)

lyric_dtd = Sequence(
    Choice(
        Sequence(
            Element(Syllabic, min_occurrence=0),
            Element(Text),
            Sequence(
                Sequence(
                    Element(Elision),
                    Element(Syllabic, min_occurrence=0),
                    min_occurrence=0
                ),
                Element(Text)
                ,
                min_occurrence=0,
                max_occurrence=None
            ),
            Element(Extend, min_occurrence=0)
        ),
        Element(Extend),
        Element(Laughing),
        Element(Humming)
    ),
    Element(EndLine, min_occurrence=0),
    Element(EndParagraph, min_occurrence=0),
    GroupReference(Editorial)
)

note_dtd = Sequence(
    Choice(
        Sequence(
            Element(Grace),
            Choice(
                Sequence(
                    GroupReference(FullNote)
                ),
                Sequence(
                    GroupReference(FullNote),
                    Element(Tie, 0, 2)
                ),
                Sequence(
                    Element(Cue),
                    GroupReference(FullNote)
                )
            )
        ),
        Sequence(
            Element(Cue),
            GroupReference(FullNote),
            GroupReference(DurationGroup)
        ),
        Sequence(
            GroupReference(FullNote),
            GroupReference(DurationGroup),
            Element(Tie, 0, 2)
        )
    ),
    Element(Instrument, 0),
    GroupReference(EditorialVoice, 0),
    Element(Type, 0),
    Element(Dot, 0, None),
    Element(Accidental, 0),
    Element(TimeModification, 0, None),
    Element(Stem, 0),
    Element(Notehead, 0),
    Element(NoteheadText, 0),
    GroupReference(Staff, 0),
    Element(Beam, 0, 8),
    Element(Notations, 0, None),
    Element(Lyric, 0, None),
    Element(Play, 0)
)


class Test(TestCase):
    def print_choices(self, dtd):
        choices = dtd.get_choices()
        for choice in choices:
            print(choice.dump())

    def test_lyric(self):
        result = [
            [[['Syllabic', 'Text', [['Elision', 'Syllabic'], 'Text'], 'Extend']], 'EndLine', 'EndParagraph',
             ['FootNote', 'Level']],
            [['Extend'], 'EndLine', 'EndParagraph', ['FootNote', 'Level']],
            [['Laughing'], 'EndLine', 'EndParagraph', ['FootNote', 'Level']],
            [['Humming'], 'EndLine', 'EndParagraph', ['FootNote', 'Level']]
        ]

        choices = [choice.get_leaves(key=lambda leaf: leaf.type_.__name__) for choice in lyric_dtd.get_choices()]
        self.assertEqual(choices, result)

    def test_note(self):
        choices = [choice.get_leaves(key=lambda leaf: leaf.type_.__name__) for choice in note_dtd.get_choices()]
        # choices = [choice.get_leaves(key=lambda leaf: leaf) for choice in note_dtd.get_dtd_choices()]

        result = [
            [[['Grace', [[['Chord', ['Pitch']]]]]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Grace', [[['Chord', ['Unpitched']]]]]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Grace', [[['Chord', ['Rest']]]]]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Grace', [[['Chord', ['Pitch']], 'Tie']]]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Grace', [['Cue', ['Chord', ['Pitch']]]]]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Cue', ['Chord', ['Pitch']], ['Duration']]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[[['Chord', ['Pitch']], ['Duration'], 'Tie']], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Grace', [[['Chord', ['Unpitched']], 'Tie']]]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type',
             'Dot', 'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Grace', [[['Chord', ['Rest']], 'Tie']]]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Grace', [['Cue', ['Chord', ['Unpitched']]]]]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type',
             'Dot', 'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Grace', [['Cue', ['Chord', ['Rest']]]]]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Cue', ['Chord', ['Unpitched']], ['Duration']]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type',
             'Dot', 'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[['Cue', ['Chord', ['Rest']], ['Duration']]], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[[['Chord', ['Unpitched']], ['Duration'], 'Tie']], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type',
             'Dot', 'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play'],
            [[[['Chord', ['Rest']], ['Duration'], 'Tie']], 'Instrument', ['FootNote', 'Level', 'Voice'], 'Type', 'Dot',
             'Accidental', 'TimeModification', 'Stem', 'Notehead', 'NoteheadText', ['StaffElement'], 'Beam',
             'Notations', 'Lyric', 'Play']
        ]
        self.assertEqual(choices, result)



from musicscore.musicxml.attributes.leveldisplay import LevelDisplay
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeSmulfAccidentalGlyphName, TypeAccidentalValue


class ComplexTypeAccidentalMark(ComplexType, TypeAccidentalValue, LevelDisplay, PrintStyle, Placement,
                                OptionalUniqueId):
    """
    An accidental-mark can be used as a separate notation or as part of an ornament. When used in an ornament, position 
    and placement are relative to the ornament, not relative to the note.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        # self.smufl = smufl

    # @property
    # def smulf(self):
    #     return self.get_attribute('smulf')
    #
    # @smulf.setter
    # def smulf(self, value):
    #     if value is None:
    #         self.remove_attribute('smulf')
    #     else:
    #         TypeSmulfAccidentalGlyphName(value)
    #         self._ATTRIBUTES.insert(0, 'smulf')
    #         self.set_attribute('smulf', value)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Id(AttributeAbstract):
    def __init__(self, id, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('id', id, "IDREF")


class PartAttributes(Id):
    """
    In either partwise or timewise format, the part element has an id attribute that is an IDREF back to a score-part
    in the part-list.
    """

    def __init__(self, id, *args, **kwargs):
        super().__init__(id=id, *args, **kwargs)



import setuptools

setuptools.setup(
    name="musicscore",
    version="1.2.7",
    author="Alex Gorji",
    author_email="aligorji@hotmail.com",
    description="generating musicxml",
    url="https://github.com/alexgorji/musicscore.git",
    packages=setuptools.find_packages(),
    install_requires=['lxml==4.3.1',
                      'quicktions==1.9'
                      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ]
)



from unittest import TestCase
from musicscore.musicxml.elements.fullnote import DisplayStep


class TestDisplayStep(TestCase):
    def setUp(self):
        self.display_step = DisplayStep(value='C')

    def test_display_step(self):
        result = '''<display-step>C</display-step>
'''
        self.assertEqual(self.display_step.to_string(), result)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class LineShape(AttributeAbstract):
    """
	The line-shape attribute distinguishes between straight and curved lines.
    """

    def __init__(self, line_shape=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('line-shape', line_shape, "TypeLineShape")



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechordflags import PercussionFlag, XFlag
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.types.complextypes.articulations import Accent
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        xml_path = path + '_test_1.xml'
        durations = [2, 1, 0.5, 0.25, 0.25, 4, 2, 3]
        sf = SimpleFormat(durations=durations)
        for chord in sf.chords:
            articulation = Accent()
            chord.add_articulation_object(articulation)

        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_2(self):
        xml_path = path + '_test_2.xml'
        articulations = ['accent', 'strong-accent', 'staccato', 'tenuto', 'detached-legato', 'staccatissimo',
                         'spiccato', 'scoop', 'plop', 'doit', 'falloff', 'breath-mark', 'caesura', 'stress', 'unstress']
        durations = len(articulations) * [1.25]
        sf = SimpleFormat(durations=durations)
        for index, chord in enumerate(sf.chords):
            chord.add_articulation(articulations[index])

        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)



from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeGroupSymbolValue


class ComplexTypeGroupSymbol(ComplexType, TypeGroupSymbolValue, Position, Color):
    """The group-symbol type indicates how the symbol for a group is indicated in the score."""

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.dtd.dtd import Sequence, Choice, Element
from musicscore.musicxml.attributes.id import Id
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import String, TypeMute, TypeSemiPitched


class Ipa(XMLElement, String):
    """The ipa element represents International Phonetic Alphabet (IPA) sounds for vocal music. String content is
    limited to IPA 2005 symbols represented in Unicode 6.0."""
    _TAG = 'ipa'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Mute(XMLElement, TypeMute):
    _TAG = 'mute'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class SemiPitched(XMLElement, TypeSemiPitched):
    _TAG = 'semi-pitched'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class OtherPlay(XMLElement):
    _TAG = 'other=play'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)
        NotImplementedError()


class ComplexTypePlay(ComplexType, Id):
    """The play type, new in Version 3.0, specifies playback techniques to be used in conjunction with the
    instrument-sound element. When used as part of a sound element, it applies to all notes going forward in score
    order. In multi-instrument parts, the affected instrument should be specified using the id attribute. When used as
    part of a note element, it applies to the current note only."""

    _DTD = Sequence(
        Choice(
            Element(Ipa),
            Element(Mute),
            Element(SemiPitched),
            Element(OtherPlay),
            min_occurrence=0,
            max_occurrence=None
        )

    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from unittest import TestCase
import musicscore.musicxml.elements.timewise as timewise
from musicscore.musicxml.elements.note import Note, Beam


class Test(TestCase):
    def setUp(self):
        self.measure = timewise.Measure('1')

    def test_number(self):
        self.assertTrue(isinstance(self.measure.number, str))
        with self.assertRaises(TypeError):
            self.measure.number = 3

    def test_non_controlling(self):
        with self.assertRaises(ValueError):
            self.measure.non_controlling = 2

        self.measure.non_controlling = 'yes'
        self.assertTrue(self.measure.non_controlling == 'yes')

    def test_add_note(self):
        part = self.measure.add_child(timewise.Part(id='bla'))
        note = part.add_child(Note())
        note.add_child(Beam('end'))
        with self.assertRaises(TypeError):
            self.measure.number = 3



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.pagemargins import ComplexTypePageMargins
from musicscore.musicxml.types.simple_type import TypeTenths


class PageHeight(XMLElement, TypeTenths):
    _TAG = 'page-height'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class PageWidth(XMLElement, TypeTenths):
    _TAG = 'page-width'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class PageMargins(ComplexTypePageMargins):
    _TAG = 'page-margins'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypePageLayout(ComplexType):
    """Page layout can be defined both in score-wide defaults and in the print element. Page margins are specified
    either for both even and odd pages, or via separate odd and even page number values. The type is not needed when
    used as part of a print element. If omitted when used in the defaults element, "both" is the default."""

    _DTD = Sequence(
        Sequence(
            Element(PageHeight),
            Element(PageWidth),
            min_occurrence=0
        ),
        Element(PageMargins, min_occurrence=0, max_occurrence=2)
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.attributes.justify import Justify
from musicscore.musicxml.attributes.printobject import PrintObject
from musicscore.musicxml.attributes.printstyle import PrintStyle


class PartNameText(PrintStyle, PrintObject, Justify):
    """
    The part-name-text attribute group is used by the part-name and part-abbreviation elements. The print-style and
    justify attribute groups are deprecated in MusicXML 2.0 in favor of the new part-name-display and
    part-abbreviation-display elements
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from musicscore.dtd.dtd import Choice, Element
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.types.complextypes.breathmark import ComplexTypeBreathMark
from musicscore.musicxml.types.complextypes.caesura import ComplexTypeCaesura
from musicscore.musicxml.types.complextypes.complextype import ComplexType, EmptyPlacement
from musicscore.musicxml.types.complextypes.emptyline import ComplexTypeEmptyLine
from musicscore.musicxml.types.complextypes.strongaccent import ComplexTypeStrongAccent


class Accent(EmptyPlacement):
    """he accent element indicates a regular horizontal accent mark."""
    _TAG = 'accent'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class StrongAccent(ComplexTypeStrongAccent):
    """The strong-accent element indicates a vertical accent mark."""
    _TAG = 'strong-accent'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Staccato(EmptyPlacement):
    """The staccato element is used for a dot articulation, as opposed to a stroke or a wedge."""
    _TAG = 'staccato'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Tenuto(EmptyPlacement):
    """The tenuto element indicates a tenuto line symbol."""
    _TAG = 'tenuto'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class DetachedLegato(EmptyPlacement):
    """The detached-legato element indicates the combination of a tenuto line and staccato dot symbol."""
    _TAG = 'detached-legato'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Staccatissimo(EmptyPlacement):
    """The staccatissimo element is used for a wedge articulation, as opposed to a dot or a stroke."""
    _TAG = 'staccatissimo'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Spiccato(EmptyPlacement):
    """The spiccato element is used for a stroke articulation, as opposed to a dot or a wedge."""
    _TAG = 'spiccato'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Scoop(ComplexTypeEmptyLine):
    """The scoop element is an indeterminate slide attached to a single note. The scoop element appears before the main
    note and comes from below the main pitch."""
    _TAG = 'scoop'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Plop(ComplexTypeEmptyLine):
    """The plop element is an indeterminate slide attached to a single note. The plop element appears before the main
    note and comes from above the main pitch."""
    _TAG = 'plop'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Doit(ComplexTypeEmptyLine):
    """The doit element is an indeterminate slide attached to a single note. The doit element appears after the main
    note and goes above the main pitch."""
    _TAG = 'plop'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Falloff(ComplexTypeEmptyLine):
    """The falloff element is an indeterminate slide attached to a single note. The falloff element appears after the
    main note and goes below the main pitch."""
    _TAG = 'falloff'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class BreathMark(ComplexTypeBreathMark):
    _TAG = 'breath-mark'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Caesura(ComplexTypeCaesura):
    """The falloff element is an indeterminate slide attached to a single note. The falloff element appears after the
    main note and goes below the main pitch."""
    _TAG = 'caesura'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Stress(EmptyPlacement):
    """The stress element indicates a stressed note."""
    _TAG = 'stress'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Unstress(EmptyPlacement):
    """The unstress element indicates an unstressed note. It is often notated using a u-shaped symbol."""
    _TAG = 'unstress'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SoftAccent(EmptyPlacement):
    """The soft-accent element indicates a soft accent that is not as heavy as a normal accent. It is often notated as
    &lt;&gt;. It can be combined with other articulations to implement the entire SMuFL Articulation Supplement
    range."""
    _TAG = 'soft-accent'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class OtherArticulation(object):
    """
    type="other-placement-text">
    The other-articulation element is used to define any articulations not yet in the MusicXML format. The smufl
    attribute can be used to specify a particular articulation, allowing application interoperability without requiring
    every SMuFL articulation to have a MusicXML element equivalent. Using the other-articulation element without the
    smufl attribute allows for extended representation, though without application interoperability.
    """
    _TAG = 'other-articulation'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        NotImplementedError()


class ComplexTypeArticulations(ComplexType, OptionalUniqueId):
    """Articulations and accents are grouped together here."""

    # _CHILDREN = [Accent, StrongAccent, Staccato, Tenuto, DetachedLegato, Staccatissimo, Spiccato, Scoop, Plop, Doit,
    #              Falloff, BreathMark, Caesura, Stress, Unstress, SoftAccent]

    _CHILDREN = [Accent, StrongAccent, Staccato, Tenuto, DetachedLegato, Staccatissimo, Spiccato, Scoop, Plop, Doit,
                 Falloff, BreathMark, Caesura, Stress, Unstress]

    _DTD = Choice(
        Element(Accent),
        Element(StrongAccent),
        Element(Staccato),
        Element(Tenuto),
        Element(DetachedLegato),
        Element(Staccatissimo),
        Element(Spiccato),
        Element(Scoop),
        Element(Plop),
        Element(Doit),
        Element(Falloff),
        Element(BreathMark),
        Element(Caesura),
        Element(Stress),
        Element(Unstress),
        # Element(SoftAccent),
        min_occurrence=0,
        max_occurrence=None
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.main_score = TreeScoreTimewise()

    def test_1(self):
        section_score_1 = TreeScoreTimewise()
        sf = SimpleFormat(durations=[1, 3, 2])
        section_score_1.set_time_signatures(durations=sum(sf.durations), barline_style='light-light')
        sf.to_stream_voice().add_to_score(section_score_1)

        section_score_2 = TreeScoreTimewise()
        sf = SimpleFormat(durations=[2, 1, 3])
        section_score_2.set_time_signatures(durations=sum(sf.durations), barline_style='heavy')
        sf.to_stream_voice().add_to_score(section_score_2)

        self.main_score.extend(section_score_1)
        self.main_score.extend(section_score_2)

        xml_path = path + '_test_1.xml'
        self.main_score.write(path=xml_path)
        TestScore().assert_template(result_path=xml_path)

    def test_2(self):
        # needed because of wrong system breaks
        section_score = TreeScoreTimewise()
        sf = SimpleFormat(durations=2 * [5, 4, 3, 2, 1])
        section_score.set_time_signatures(durations=sum(sf.durations), barline_style='light-light')
        sf.to_stream_voice().add_to_score(section_score)
        # section_xml_path = path + '_test_2_section.xml'
        # section_score.write(path=section_xml_path)

        self.main_score.extend(section_score)
        main_xml_path = path + '_test_2_main.xml'
        self.main_score.write(path=main_xml_path)

        TestScore().assert_template(result_path=main_xml_path)



from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeNoteTypeValue, TypeSymbolSize


class ComplexTypeNoteType(ComplexType, TypeNoteTypeValue):
    """The note-type type indicates the graphic note type. Values range from 1024th to maxima. The size attribute
    indicates full, cue, grace-cue, or large size. The default is full for regular notes, grace-cue for notes that
    contain both grace and cue elements, and cue for notes that contain either a cue or a grace element, but not both.
    """
    _ATTRIBUTES = []

    def __init__(self, tag, size=None, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.size = size

    @property
    def size(self):
        return self.get_attribute('size')

    @size.setter
    def size(self, value):
        if value is None:
            self.remove_attribute('size')
        else:
            TypeSymbolSize(value)
            self._ATTRIBUTES.insert(0, 'size')
            self.set_attribute('size', value)
